/* FILE: grid.cc           -*-Mode: c++-*-
 *
 * This file contains class definitions for micromagnetic grid classes.
 *
 */

/* Revision history:
 *  7-Apr-96: Removed last vestiges of g2d file format.
 *  9-Jul-96: Added support for maximum angle difference determination.
 *  8-Jul-96: Revamped ODE solver.  Introduced 4th order Runge-Kutta
 *            solver with adaptive stepsize.
 * 30-Jun-96: Added Array2D<MagElt> support.
 */

#include <ctype.h>
#include <string.h>

#include "nb.h"

#include "demag.h"
#include "grid.h"
#include "maganis.h"
#include "magelt.h"
#include "vecio.h"
#include "vector.h"

/* End includes */

// Define UNIT_MAG to get speed-ups from assuming all magnetizations
// are unit vectors
#undef UNIT_MAG
#undef WEAK_BOUNDARY // 0.5  // Set boundary element reduced moments to this
                           // #undef or set to 0.0 to get sharp boundaries.
#undef EDGE_ADJUST // 1.0 // Alternate form of WEAK_BOUNDARY; Affects demag
                    // only, and only works with RECTANGULAR sample shapes!!!
#undef ROUGH_BOUNDARY
    // If #defined, then each boundary moment is independently multiplied
    // by an amount drawn from a uniform distribution on [0,1].

// ODE parameters
const double Grid2D::GyRatio=2.21e5; // Gyromagnetic ratio, m/A.s
const double Grid2D::DampCoef=0.5;   // Damping coefficient, dimensonless
// The damping coefficient can be obtained via FMR studies as
//  DampCoef=(Delta H).GyRatio/(2.(fmr_driving_frequency)).
//  'Delta H' here is the FMR line width.  In the literature,
//  GyRatio is denoted by \gamma, fmr_driving_frequency is \omega_0,
//  and DampCoef is often written \alpha or \alpha/(1+\alpha).
// Realistic values for DampCoef are in the range [.004,.15], but for
//  fast convergence of the ODE, DampCoef=0.5 seems close to optimal.
// dm/dt = (-GyRatio.Ms)(mxh)-(DampCoef.GyRatio.Ms)(mx(mxh))
//  This program actually solves RHS=(-1/DampCoef)(mxh)-mx(mxh).
//  The routine Grid2D::GetTimeStep() does the conversion from StepSize
//  to actual time step in seconds, via
//              time_step = StepSize/(DampCoef.GyRatio.Ms)
//  See also PRECESSION_RATIO below.

// #define exactly one of the following to select the coefficient
// to use with the internal demag routine.
#undef CYL_COEFS      // This is the one you get if you don't #define any
#undef SQUARE_COEFS
#define PAPER_COEFS

// Miscellaneous #define's
#undef OLD_CODE
#undef VERBOSE
#undef PRECESSION   // Define this to turn on precession term
#define PRECESSION_RATIO (-1./Grid2D::DampCoef)
          // Relative size of precession term. See above discussion
          // about DampCoef and GyRatio.

#undef CODE_CHECKS

// ODE Solvers; Define exactly one of the following
#undef EULER_ODE       // 1st order Euler
#define PREDICTOR_2_ODE // 2nd order predictor
#undef RUNGE_KUTTA_ODE  // 4th order Runge-Kutta
                        // (cf. MaxTorqueStep note below)
#undef RUNGE_KUTTA_FAST // Define this *in addition to RUNGE_KUTTA_ODE*
                         // to use approximate field values (i.e.,
                         // hFastUpdate) for internal ODE steps.

// Error control
#ifndef RUNGE_KUTTA_ODE
const double MaxTorqueStep  =  0.09; // StepSize*maxtorque restricted to
     // being less than this amount.  This will roughly limit magnetization
     // shifts to < atan(0.09) --- about 5 degrees.
#else  // Current Runge-Kutta solver takes double step
const double MaxTorqueStep  =  0.18;
#endif
const double AllowedSpinError =  1e-3;  // Allowed error per spin
/// Affects ODE solvers.

#ifdef STOPWATCH_ON
#define STOPWATCH_START(watch) watch.Start();
#define STOPWATCH_STOP(watch)  watch.Stop();
#else // !STOPWATCH_ON
#define STOPWATCH_START(watch)
#define STOPWATCH_STOP(watch)
#endif // STOPWATCH_ON

// Grid (mesh) file magic cookie string
// const char GridFileMagic03[8] = "G2DFil\x03";   OBSOLETE
// const double GridFileEndianTest = 1.234;        OBSOLETE
// The bit pattern for 1.234 on a little endian machine (e.g.,
// Intel's x86 series) should be 58 39 B4 C8 76 BE F3 3F, while
// on a big endian machine it is 3F F3 BE 76 C8 B4 39 58.  Examples
// of big endian machines include ones using the MIPS R4000 and R8000
// series chips, for example SGI's.


/////////////////////////// Grid2D Code /////////////////////////////////

// Coordinate transformations.  The 2D grid/solver uses x+z in-plane
// (width and height), with y pointing into the screen (thickness).
//  For reference this is called the "xzy" coordinate system.  All
// outside OOMMF code (including the MIF class) uses x+y in-plane
// (width and height), with z pointing out of the screen (thickness).
// This is referred to as the "xyz" coordinate system.
void Grid2D::ConvertXyzToXzy(ThreeVector &vec)
{
  double temp=vec[1];  vec[1]=-vec[2];  vec[2]=temp;
}

void Grid2D::ConvertXzyToXyz(ThreeVector &vec)
{
  double temp=vec[1];  vec[1]=vec[2];  vec[2]=-temp;
}


double Grid2D::GetTimeStep(void)
{ // Returns size of last step, in seconds
  return StepSize/(GyRatio*DampCoef*Ms);
}

void Grid2D::Allocate(int xcount,int zcount)
{ // NOTE: This routine requires that Grid2d::ExternalDemag be initialized,
  //       since *ExternalDemag->init is called.
  int i;

  Nx=xcount; Nz=zcount;
  if(DoDemag) ThreeVector::Alloc2DArray(Nx,Nz,hdemag);
  else        hdemag=(ThreeVector**)NULL;
  ThreeVector::Alloc2DArray(Nx,Nz,h);
  ThreeVector::Alloc2DArray(Nx,Nz,torque);
  ThreeVector::Alloc2DArray(Nx,Nz,torque0);
  ThreeVector::Alloc2DArray(Nx,Nz,torque1);
  ThreeVector::Alloc2DArray(Nx,Nz,h0);
  ThreeVector::Alloc2DArray(Nx,Nz,h1);

  // Allocate MagElt arrays
  m0.Allocate(Nx,Nz);
  m1.Allocate(Nx,Nz);
  m2.Allocate(Nx,Nz);
  m.Allocate(Nx,Nz);
  moment_sum=Nx*Nz;  // Note: This will be recalculated after
  /// the sample geometry is set.

  // Allocate demag calculation arrays
  if(DoDemag) {
    if(ExternalDemag->calc==NULL) { // Default to internal routine
      // Allocate demagcoef matrices
      if(!(ADemagCoef=new double*[Nx]))
	PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
      if(!(CDemagCoef=new double*[Nx]))
	PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
      for(i=0;i<Nx;i++) {
	if(!(ADemagCoef[i]=new double[Nz]))
	  PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
	if(!(CDemagCoef[i]=new double[Nz]))
	  PlainError(1,"Error in Grid2D::Allocate: %s",ErrNoMem);
      }
    }
    else { // Use external routine
      (*ExternalDemag->init)(Nx,Nz,ExternalDemag->paramcount,
			     ExternalDemag->params);
    }
  } // fi(DoDemag)
}


int Grid2D::FindNeighbors(int nsize,Array2D<MagElt> &me,MagEltLink men[],
			       int i,int k)
{ // NOTE: Currently does only 4 and 8 neighborhoods.
  // NOTE 2: Enforces Neuman bdry conditions along edges of gridded
  //       region.  Hmm, you may get different results if the sample edge
  //       is not the same as the gridded region edge (e.g., by setting
  //       |m|=0 inside gridded region to, for example, round corners.)
  //       -mjd 5-Dec-1996.
#ifdef CODE_CHECKS
  if(nsize!=4 && nsize!=8)
    PlainError(1,"Error in Grid2D::FindNeighbors: "
	     "Unsupported neighborhood size (%d) request",nsize);
  if(me.GetSize()<1 || men==(MagEltLink*)NULL) 
    PlainError(1,"Error in Grid2D::FindNeighbors: Null pointer received");
  if(i<0 || i>=Nx || k<0 || k>=Nz)
    PlainError(1,"Error in Grid2D::FindNeighbors: "
	     "Requested index elt (%d,%d) out-of-bounds (0-%d,0-%d)",
	     i,k,Nx,Nz);
#endif
  hValid=torqueValid=energyValid=0;
  if(me[i][k].GetMoment()<=0.) return 0; // Zero moment elements
  /// don't get connected.

  static int ioff[8]={ 1, 1, 0,-1,-1,-1, 0, 1};
  static int koff[8]={ 0, 1, 1, 1, 0,-1,-1,-1};
  int n,offset,offstep,it,kt,new_ngbrcnt;
  if(nsize==4) offstep=2;  else offstep=1;
  for(n=offset=new_ngbrcnt=0;n<nsize;n++,offset+=offstep) {
    it=i+ioff[offset];
    kt=k+koff[offset];

    // Enforce Neumann bdry conds.,  \part(m)/\part(n)=0,
    // at grid edges
    if(it<0)        it=0;
    else if(it>=Nx) it=Nx-1;
    if(kt<0)        kt=0;
    else if(kt>=Nz) kt=Nz-1;
    if(it==i && kt==k) continue;

    // Handle irregular boundaries
    if(me[it][kt].GetMoment()<=0.) {
      if(it==i || kt==k) continue;  // Set up alternate connection
      /// only for diagonal links

      // See which of the nearest neighbors is non-zero
      int itt=it;  int ktt=kt;
      if(me[i][kt].GetMoment()>0.) itt=i;
      if(me[it][k].GetMoment()>0.) ktt=k;

      if(itt==it && ktt==kt) continue; // Neither; Skip link
      if(itt==i  && ktt==k)  continue; // Both; Skip link

      it=itt; kt=ktt;
    }

#ifdef CODE_CHECKS
    if((it==i && kt==k) || me[it][kt].GetMoment()<=0.)
      PlainError(1,"Error in Grid2D::FindNeighbors: "
	       "Programming error: Bad neighbor link");
#endif // CODE_CHECKS

    // Add neighbor
    men[new_ngbrcnt].weight=1.0;
    men[new_ngbrcnt++].node=&(me[it][kt]);    
  }
  return new_ngbrcnt;
}
#ifndef UNIT_MAG
void Grid2D::InitRectangle(Nb_DString &)
{ // Default part geometry
  return; // Nothing to do
}
void Grid2D::InitEllipse(Nb_DString &)
{ // Zero's moments outside ellipse defined by Nx,Nz
  if(Nx<2 || Nz<2) return; // Nothing to do
  int i,k; double x,z;
  for(i=0;i<Nx;i++) {
    x= double(2*i)/double(Nx-1)-1.0;
    for(k=0;k<Nz;k++) {
      z= double(2*k)/double(Nz-1)-1.0;
      if(OC_SQ(x)+OC_SQ(z)>1+sqrt(DBL_EPSILON)) m[i][k].SetMoment(0.0);
    }
  }
  hValid=torqueValid=energyValid=0;
}

static char* PpmStrtok(char* str)
{ // Utility function used by Grid2D::Mask(Nb_DString).
  // Basically a replacement for strtok() that ignores
  // comments.
  static char* cpstart="";
  if(str!=NULL) cpstart=str;
  char *cp=cpstart;
  // Skip leading whitespace and comments
  while(*cp!='\0') {
    while(*cp!='\0' && isspace(*cp)) cp++;
    if(*cp!='#') break;  // Non-comment
    if((cp=strchr(cp,'\n'))==NULL) break; // Skip past comment to eol.
    cp++;  // Skip over end-of-line.
  }
  if(cp==NULL || *cp=='\0') {
    cpstart="";
    return NULL;
  }
  char* cpnewstart=cp;  // Record first non-whitespace non-comment char

  // Find end of data field
  while(*cp!='\0' && *cp!='#' && !isspace(*cp)) cp++;

  // Mark end of field with null, but preserve comment characters
  // for next pass.
  if(*cp=='#' && *(cp+1)!='\n' && *(cp+1)!='\0') *(cp+1)='#';
  if(*cp!='\0') {
    *cp='\0';
    cpstart=cp+1;
  } else {
    cpstart="";
  }
  return cpnewstart;
}

void Grid2D::Mask(Nb_DString &filepoint)
{ // Zeros moments if not black
  if(Nz<1 || Nx<1) return; // Nothing to fill (empty grid).
  // Read data and store into "slurp" string
  Oc_AutoBuf ab1,ab2,ab3;
  Tcl_Interp* interp=Oc_GlobalInterpreter();
  if(Tcl_VarEval(interp,ab1("SlurpImageFile {"),
                 ab2(filepoint.GetStr()),ab3("} __slurped_data"),
                 (char *)NULL)!=TCL_OK) {
    PlainWarning("Warning reading mask file %s: %s (Grid2D::Mask)",
                 filepoint.GetStr(),interp->result);
    return;
  }
  char *slurp=Tcl_GetVar(interp,ab1("__slurped_data"),0);
  if(slurp==NULL) {
    PlainError(1,"Error in Grid2D::Mask(%s):"
               " Unable to access slurped data.",filepoint.GetStr());
  }
  char *cp=PpmStrtok(slurp);
  if(cp==NULL || strcmp(cp,"P3")!=0) {
    PlainWarning("Warning in Grid2D::Mask:"
                 " Can't process mask %s because PPM id string"
                 " is %s != P3",filepoint.GetStr(),(cp==NULL)?"null":cp);
    Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
    return;
  }
  if((cp=PpmStrtok(NULL))==NULL) {
    PlainWarning("Warning in Grid2D::Mask:"
                 " Premature end of file in mask file %s",
		 filepoint.GetStr());
    Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
    return;
  }
  int picwidth=atoi(cp);
  if((cp=PpmStrtok(NULL))==NULL) {
    PlainWarning("Warning in Grid2D::Mask:"
                 " Premature end of file in mask file %s",
		 filepoint.GetStr());
    Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
    return;
  }
  int picheight=atoi(cp);
  if((cp=PpmStrtok(NULL))==NULL) {
    PlainWarning("Warning in Grid2D::Mask:"
                 " Premature end of file in mask file %s",
		 filepoint.GetStr());
    Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
    return;
  }
  int maxval=atoi(cp);
  if(picheight<1 || picwidth<1) {
    PlainWarning("Warning in Grid2D::Mask(Nb_DString %s): "
               "Empty image (%d x %d)",filepoint.GetStr(),
               picwidth,picheight);
    Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
    return;
  }

  // Process slurped data
  int i,j;
  int vpp=3;    // 3 values per pixel
  char *rowbuf=new char[vpp*picwidth];
  double x,xstep,r,rstep,z,zstep;
  xstep=double(picwidth)/double(Nx);
  if(Nz>=picheight) {
    // More simulation rows than picture rows
    zstep=1.; rstep=double(picheight)/double(Nz);
  } else {
    // More picture rows than simulation rows
    rstep=1.; zstep=double(Nz)/double(picheight);
  }
  for(j=0,r=-rstep/2.,z=-zstep/2. ; j<OC_MAX(Nz,picheight);
      j++,r+=rstep,z+=zstep) {
    if(r<floor(r+rstep)) {
      // Parse 1 image row
      if(floor(r+rstep)<picheight) { // Safety check
        for(i=0;i<vpp*picwidth;i++) {
	  if((cp=PpmStrtok(NULL))==NULL) {
	    PlainWarning("Warning in Grid2D::Mask:"
			 " Premature end of file in mask file %s",
			 filepoint.GetStr());
	    Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
	    return;
	  }
	  rowbuf[i]=atoi(cp);
	}
      }
    }
    if(z<floor(z+zstep)) {
      // Put current row into grid
      int gridk=Nz-1-int(floor(z+zstep));
      if(gridk<0) gridk=0;  // Safety
      for(i=0,x=0.;i<Nx;i++,x+=xstep) {
        int offset=int(floor(x));
        if(offset>picwidth-1) offset=picwidth-1;
        offset*=vpp;
        for(int p=0;p<vpp;p++) {
          if(rowbuf[offset++]!=0) {
            m[i][gridk].SetMoment(0.);
            break;
          }
        }
      }
    }
  }
  delete[] rowbuf;
  Tcl_UnsetVar(interp,ab1("__slurped_data"),0);
}

#ifdef OLD_CODE
void Grid2D::Mask(Nb_DString &filepoint)
{ // Zeros moments if not black
#define Maxline 256 
  static char buf[Maxline];
  int i,k,hapval=0,ppmtype=0;
  FILE* mp;

  Nb_TclNativeFilename(filepoint);
  mp=Nb_FOpen(filepoint.GetStr(),"rb");
  if(mp==NULL) {
    PlainError(1,"Error in Grid2D::Mask: Unable to open file: %s",
	       filepoint.GetStr());
  }

  while(hapval<=2) {    
    if(fgets(buf,Maxline,mp)==NULL)
      PlainError(1,"Error in Grid2D::Mask: Error reading file: %s",
		 filepoint.GetStr());
    if((strstr(buf,"#")==NULL)&&(strspn(buf," \t\n")!=strlen(buf)))  {
      if((hapval==0)&&(strstr(buf,"P3")!=NULL)) ppmtype=3;
      else if ((hapval==0) &&
	       ((strstr(buf,"P6")!=NULL)|| (strstr(buf,"P4")!=NULL)))
	ppmtype=6;
      else if (hapval==0)
	PlainError(1,"Error in Grid2D::Mask: Not a valid ppm file: %s",
		   filepoint.GetStr());
      hapval++;
    } 
  }

  if (ppmtype==3) {
    int Red,Green,Blue;
    for(k=Nz-1;k>=0;k--)  for(i=0;i<Nx;i++){
      if (fscanf(mp,"%i %i %i",&Red,&Green,&Blue)==EOF)
	PlainError(1,"Error in Grid2D::Mask(Nb_DString filepoint) "
		 "PPM size does not match grid size in: %s",
		 filepoint.GetStr());
      else
        if(Red+Green+Blue) m[i][k].SetMoment(0.0); 
    }
  }
  else {
    int Check=0,*intbuf;
    intbuf=&Check;
    for(k=Nz-1;k>=0;k--)  for(i=0;i<Nx;i++){
      if(feof(mp))
        PlainError(1,"Error in Grid2D::Mask(Nb_DString filepoint) "
		   "PPM size does not match grid size in: %s",
		   filepoint.GetStr());
      else {
	fread((char *)intbuf,3,1,mp);
	if(Check) m[i][k].SetMoment(0.0);    
      }
    }
  }
  fclose(mp);
}
#endif // OLD_CODE

void Grid2D::InitOval(Nb_DString &inprop)
{ // Rounds corners off Nx by Nz rectangle by zeroing
  // moments outside of quarter circles of radius...
  double rounded_proportion=atof(inprop.GetStr());
  
  double radius=0.5*rounded_proportion*OC_MIN(Nx,Nz);

  // Corner rounding circle center coordinates
  double cx1,cx2,cz1,cz2;
  cx1=radius-0.5;   cx2=Nx-0.5-radius;
  cz1=radius-0.5;   cz2=Nz-0.5-radius;
  if(cx1>cx2 || cz1>cz2)
    PlainError(1,"Error in Grid2D::InitOval: %s",ErrProgramming);

  // Round lower left corner
  int i,k;   double xdistsq,zdistsq,radsq=OC_SQ(radius);
  for(i=0;i<(int)floor(cx1);i++) {
    xdistsq=cx1-i; xdistsq*=xdistsq;
    for(k=0;k<(int)floor(cz1);k++) {
      zdistsq=cz1-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+sqrt(DBL_EPSILON)) m[i][k].SetMoment(0.0);
    }
  }
  // Round lower right corner
  for(i=(int)ceil(cx2);i<Nx;i++) {
    xdistsq=cx2-i; xdistsq*=xdistsq;
    for(k=0;k<(int)floor(cz1);k++) {
      zdistsq=cz1-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+sqrt(DBL_EPSILON)) m[i][k].SetMoment(0.0);
    }
  }
  // Round upper right corner
  for(i=(int)ceil(cx2);i<Nx;i++) {
    xdistsq=cx2-i; xdistsq*=xdistsq;
    for(k=(int)floor(cz2);k<Nz;k++) {
      zdistsq=cz2-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+sqrt(DBL_EPSILON)) m[i][k].SetMoment(0.0);
    }
  }
  // Round upper left corner
  for(i=0;i<(int)floor(cx1);i++) {
    xdistsq=cx1-i; xdistsq*=xdistsq;
    for(k=(int)floor(cz2);k<Nz;k++) {
      zdistsq=cz2-k; zdistsq*=zdistsq;
      if(xdistsq+zdistsq>radsq+sqrt(DBL_EPSILON)) m[i][k].SetMoment(0.0);
    }
  }

  hValid=torqueValid=energyValid=0;
}

void Grid2D::SetBoundaryMoments(double momentsize)
{ // Sets all boundary magelt's moment magnitudes to momentsize
  // Note: This subroutine will likely change which elements
  //       answer "1" to MagElt::IsBoundary().
  int i,k,bdrycount=0;
  MagElt** meptr=new MagElt *[Nx*Nz];
  if(!meptr) 
    PlainError(1,"Error in Grid2D::SetBoundaryMoments(double): %s",ErrNoMem);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    if(m[i][k].GetMoment()>0. && m[i][k].IsBoundary())

      meptr[bdrycount++]=&m[i][k];
  }
  for(i=0;i<bdrycount;i++) meptr[i]->SetMoment(momentsize);
  delete meptr;
  hValid=torqueValid=energyValid=0;
}

void Grid2D::RoughenBoundaryMoments()
{ // Multiplies each boundary moment (separately) by an amount
  // drawn from a uniform distribution on [0,1].
  int i,k,bdrycount=0;
  MagElt** meptr=new MagElt *[Nx*Nz];
  if(!meptr) {
    PlainError(1,"Error in Grid2D::RoughenBoundaryMoments(double): %s",
	       ErrNoMem);
  }
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    if(m[i][k].GetMoment()>0. && m[i][k].IsBoundary())
      meptr[bdrycount++]=&m[i][k];
  }
  for(i=0;i<bdrycount;i++) {
    meptr[i]->SetMoment(meptr[i]->GetMoment()*UnifRand());
  }
  delete meptr;
  hValid=torqueValid=energyValid=0;
}

#endif // UNIT_MAG


void Grid2D::Initialize(MIF &mif,MagEltAnisType new_magelt_anistype)
{
  double new_Ms=mif.GetMs();
  double new_K1=mif.GetK1();
  double new_exchange_stiffness=mif.GetA();
  double new_stepsize=mif.GetInitialIncrement();
  double new_cellsize=mif.GetCellSize();

  int i,k;
  double dtemp;

  energy=DBL_MAX;  // Initialize to big number
  exch_energy=anis_energy=demag_energy=zeeman_energy=energy;

  Ms                 = new_Ms;
  K1                 = new_K1;
  exchange_stiffness = new_exchange_stiffness;
  magelt_anistype    = new_magelt_anistype;
  MagElt::SetupConstants(new_Ms,new_exchange_stiffness,
			 new_cellsize,new_magelt_anistype);
  StepSize=OC_MAX(REAL8m_EPSILON,new_stepsize);
  InitialStepSize=NextStepSize=StepSize0=StepSize;
  hUpdateCount=0; // Count of calls to hUpdate
  ODEIterCount=0; // ODE history state count
  hValid=torqueValid=energyValid=0;

  // Setup internal demag function, if necessary
  // (see also Grid2D::Allocate())
  if(ExternalDemag->calc==(DF_Calc *)NULL && DoDemag) {
    ADemagCoef[0][0]=CDemagCoef[0][0]=0.0;
#if defined PAPER_COEFS
    SelfDemag= -0.5;
#elif defined SQUARE_COEFS
    SelfDemag= -0.5;
#else // CYL_COEFS
    SelfDemag=(-PI/6);
#endif
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      if(i!=0 || k!=0) {
#if defined PAPER_COEFS // Use MSZ coefficients (APS Vol 49, #22, p15745-15752)
	if((dtemp=OC_SQ(2*i+1)+4*OC_SQ(k))<0)
	  PlainError(1,"Error in Grid2D::Initialize: %s","Integer overflow");
	ADemagCoef[i][k]=atan2(4*i-2,OC_SQ(2*i-1)+4*OC_SQ(k)-1)
	  -atan2(4*i+2,OC_SQ(2*i+1)+4*OC_SQ(k)-1);
	dtemp=4*OC_SQ(OC_SQ(i))+8*OC_SQ(i)*OC_SQ(k)+4*OC_SQ(OC_SQ(k))+1;
	CDemagCoef[i][k]=0.5*log((dtemp+8*i*k)/(dtemp-8*i*k));
	ADemagCoef[i][k]/=2*PI; // I don't know if this is the proper
	CDemagCoef[i][k]/=2*PI; // normalization or not, but it looks good.
#elif defined SQUARE_COEFS
	dtemp=(2*k+1)/sqrt(OC_SQ(2*i-1)+OC_SQ(2*k+1));
	dtemp-=(2*k-1)/sqrt(OC_SQ(2*i-1)+OC_SQ(2*k-1));
	ADemagCoef[i][k]=2*dtemp/(2*i-1);
	dtemp=(2*k+1)/sqrt(OC_SQ(2*i+1)+OC_SQ(2*k+1));
	dtemp-=(2*k-1)/sqrt(OC_SQ(2*i+1)+OC_SQ(2*k-1));
	ADemagCoef[i][k]-=2*dtemp/(2*i+1);
	CDemagCoef[i][k]=2./sqrt(OC_SQ(2*i-1)+OC_SQ(2*k-1))
	                -2./sqrt(OC_SQ(2*i-1)+OC_SQ(2*k+1))
	                -2./sqrt(OC_SQ(2*i+1)+OC_SQ(2*k-1))
	                +2./sqrt(OC_SQ(2*i+1)+OC_SQ(2*k+1));
#else // CYL_COEFS
	if((dtemp=(i*i+k*k))<1)
	  PlainError(1,"Error in Grid2D::Initialize: Integer overflow");
	dtemp=2*PI*OC_SQ(dtemp); // Denominator-Area consider. cancel cellsize
	ADemagCoef[i][k]=(i*i-k*k)/dtemp;
	CDemagCoef[i][k]=2.0*(i*k)/dtemp;
#endif // Coefficient selection
      }
    }
  }

  // Set initial H field to 0
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    h[i][k][0]=h[i][k][1]=h[i][k][2]=0.0;
    h0[i][k][0]=h0[i][k][1]=h0[i][k][2]=0.0;
    h1[i][k][0]=h1[i][k][1]=h1[i][k][2]=0.0;
    if(DoDemag) hdemag[i][k][0]=hdemag[i][k][1]=hdemag[i][k][2]=0.0;
  }

  // Set initial torques to 0
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    torque1[i][k][0]=torque1[i][k][1]=torque1[i][k][2]=0.0;
    torque0[i][k][0]=torque0[i][k][1]=torque0[i][k][2]=0.0;
    torque[i][k][0]=torque[i][k][1]=torque[i][k][2]=0.0;
  }

  hValid=torqueValid=energyValid=0;  // Safety;  Should be 0 from above
}

int Grid2D::InitMagnetization()
{ // Note: the maginit functions work in the OOMMF xy+z coordinate
  // system, so after initialization we have to do a coordinate
  // change on the magnetization vectors.
  if(MagInitArgs.argc<1) return 1; // Safety
  MI_Function* InitM=MagInit::LookUp(MagInitArgs.argv[0]);
  if(InitM==NULL) return 2;        // Safety
  (*InitM)(Nx,Nz,m,&(MagInitArgs.argv[1]));
  for(int i=0;i<Nx;i++) for(int k=0;k<Nz;k++) {
    ConvertXyzToXzy(m[i][k]);
    m[i][k].ValidateMoment(); // Protect against poorly designed
    /// maginit functions.
  }
  return 0;
}

void Grid2D::Reset()
{ // The calling routine is responsible for resetting hApplied
  UnifRand(randomizer_seed);
  InitMagnetization();
  StepSize0=StepSize=NextStepSize=InitialStepSize;
  ResetODE();
}

Grid2D::Grid2D(MIF &mif)
{
  STOPWATCH_START(sw_proc_total);
  if(!mif.IsValid()) {
    PlainError(1,"Error in Grid2D::Grid2D: Invalid mif object.");
  }
  dfparams=(char **)NULL; // Safety;
  randomizer_seed=mif.GetRandomizerSeed();
  UnifRand(randomizer_seed);

  int xcount,zcount,i,k;
  MagEltAnisType new_magelt_anistype=meat_uniaxial; // Safety
  moment_sum=1.0;  // Safety

  // Get basic info
  if((cellsize=mif.GetCellSize())<=0.0)
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	     "Illegal cell size (%f meters)",cellsize);
  xcount=(int)OC_ROUND(mif.GetPartWidth()/cellsize);
  zcount=(int)OC_ROUND(mif.GetPartHeight()/cellsize);
  xmax=xcount*cellsize;
  zmax=zcount*cellsize;
  ymax=mif.GetPartThickness();

  // Do demag initialization
  if(!(dfparams=new char*[OC_MAX(2,MAXDFPARAMS)])
     || !(dfparams[0]=new char[MAXDFNAMELENGTH]))
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): %s",ErrNoMem);
  strncpy(dfparams[0],mif.GetDemagName(),MAXDFNAMELENGTH);
  *(dfparams[1]=(char *)strchr((const char *)dfparams[0],'\0')+1)='\0';
  if(DF_LookUpName(dfparams[0],&ExternalDemag)!=0) // No match!
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	     "No demag calculation routine by the name ->%s<- found",
	     dfparams[0]);
  if(Nb_StrCaseCmp(ExternalDemag->name,"NoDemag")==0) DoDemag=0;
  else                                                 DoDemag=1;
  if(ExternalDemag->paramcount<1) ExternalDemag->params=(char**)NULL;
  else {
    ExternalDemag->params=dfparams;
    if(ExternalDemag->paramcount>1)
      sprintf(dfparams[1],"%.15g",mif.GetPartThickness()/cellsize);
  }

  // Determine magelt_anistype and anisotropy directions
#if (VECDIM != MIFVECDIM)
#error VECDIM and MIFVECDIM must be equal
#endif
  switch(mif.GetAnisType())
    {
    case mif_uniaxial: new_magelt_anistype=meat_uniaxial; break;
    case mif_cubic:    new_magelt_anistype=meat_gencubic; break;
    default: PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
			"Unknown Anisotropy Type in MIF");
	     break;
    }
  double dirA_arr[VECDIM];  ThreeVector dirA(dirA_arr);
  double dirB_arr[VECDIM];  ThreeVector dirB(dirB_arr);
  double dirC_arr[VECDIM];  ThreeVector dirC(dirC_arr);
  Convert(dirA,mif.GetAnisDir(1));
  ConvertXyzToXzy(dirA);  // Convert to internal coords
  if(fabs(dirA.MagSq()-1)>8*sqrt(REAL8m_EPSILON))
    PlainError(1,"Error in Grid2D::Grid2D(MIF &):"
	       " Non-unit anisotropy direction (1).");
  if(new_magelt_anistype!=meat_uniaxial) { // Some cubic type
    Convert(dirB,mif.GetAnisDir(2));
    ConvertXyzToXzy(dirB);  // Convert to internal coords
    if(fabs(dirB.MagSq()-1)>8*sqrt(REAL8m_EPSILON))
      PlainError(1,"Error in Grid2D::Grid2D(MIF &):"
		 " Non-unit anisotropy direction (2).");
    double dot=dirA*dirB;
    if(fabs(dot)>8*sqrt(REAL8m_EPSILON))
      PlainError(1,"Error in Grid2D::Grid2D(MIF &):"
		 " Crystal axes not perpendicular.");
    dirC=dirA^dirB;
    if(fabs(dirC.MagSq()-1)>8*sqrt(REAL8m_EPSILON))
      PlainError(1,"Error in Grid2D::Grid2D(MIF &):"
		 " Non-unit anisotropy direction (3).");
    if(dirA[0]==1.0 && dirB[1]==1.0)
      new_magelt_anistype=meat_cubic;  // Standard cubic anisotropy
  }

  // Allocate arrays
  Allocate(xcount,zcount);

  // Initialize program constants
  Initialize(mif,new_magelt_anistype);

  // Setup applied field.  The SetCoords() call must be done
  // *after* Nx & Nz are setup, which is currently done in Allocate.
  // Perhaps the SetCoords() call should be moved there as well?
  ArgLine zargs=mif.GetZeemanArgs();
  if(zargs.argc<1) {
    // Applied field not specified.  Default to uniform field.
    const char *temp[1]={ "uniform" };
    hApplied=Zeeman::MakeZeeman(1,(const char **)temp);
  } else {
    hApplied=Zeeman::MakeZeeman(zargs.argc,(const char **)zargs.argv);
  }
  hApplied->SetCoords(this,Nx,Nz);
  hApplied->SetNomFieldVec3D(Ms,Vec3D(0.,0.,0.)); // Init to 0 field.

  // Setup anisotropy axes inside MagElt's.  Anisotropy arrays should
  // be unchanging, so they can be shared across all the magelt arrays.
  ANIS_INIT_FUNC* aifp=AnisInitList.GetAnisInitFunc(mif.GetAnisInitName());
  if(aifp==NULL)
    PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	       "Unknown AnisInitFunc: ->%s<-",mif.GetAnisInitName());
  ArgLine ai=mif.GetAnisInitArgs();
  (*aifp)(ai.argc,(const char* const*)ai.argv,  // Cast to keep VC++ happy
	  Nx,Nz,magelt_anistype,K1,dirA,dirB,dirC,
	  m,anisa,anisb,anisc);
  double *aptr=NULL,*bptr=NULL,*cptr=NULL;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m[i][k].GetAnisDirs(aptr,bptr,cptr);   // Copy anis pointers from m
    m0[i][k].InitAnisDirs(aptr,bptr,cptr); // to m0,
    m1[i][k].InitAnisDirs(aptr,bptr,cptr); // to m1,
    m2[i][k].InitAnisDirs(aptr,bptr,cptr); // and to m2,
  }

#ifndef UNIT_MAG
  // Setup sample geometry.  NOTE: We need to setup the geometry (by
  //  setting some magelt moments to 0) _before_ calling the neighborhood
  //  constructor routine, because the latter needs to know where the
  //  boundaries are in order to get proper boundary conditions.
  // NOTE 2: These geometry initialization routines must not
  //  use the m->ngbr arrays (obviously).
   MifGeometryType mgt; Nb_DString geomparam;
   mgt=mif.GetGeometry(geomparam);
  if(mgt==mif_rectangle)    InitRectangle(geomparam);
  else if(mgt==mif_ellipse) InitEllipse(geomparam);
  else if(mgt==mif_oval)    InitOval(geomparam);
  else if(mgt==mif_mask)    Mask(geomparam);
  else PlainError(1,"Error in Grid2D::Grid2D(MIF &): Part geometry not specified.");
#ifdef WEAK_BOUNDARY
  if(WEAK_BOUNDARY>0.0) SetBoundaryMoments(WEAK_BOUNDARY);
#endif // WEAK_BOUNDARY
#ifdef ROUGH_BOUNDARY
  RoughenBoundaryMoments();
#endif // ROUGH_BOUNDARY
#endif // !UNIT_MAG

  // Initialize m and copy data (vector, moment and K1) into m0, m1,
  // and m2.  Note that _after_ this all magelt's have their moment
  // magnitude set, so then one may set up the nbhd structure.
  //  Also, the maginit functions work in the OOMMF xy+z coordinate
  // system, so after initialization we have to do a coordinate
  // change on the magnetization vectors.
  MagInitArgs=mif.GetMagInitArgs();
  InitMagnetization();
  moment_sum=0.0;
  for(i=0;i<Nx;i++) {
    double moment_column_sum=0.0;
    for(k=0;k<Nz;k++) {
      moment_column_sum+=fabs(m[i][k].GetMoment());
      m0[i][k].CopyData(m[i][k]);
      m1[i][k].CopyData(m[i][k]);
      m2[i][k].CopyData(m[i][k]);
    }
    moment_sum+=moment_column_sum;
  }
  if(moment_sum<REAL8m_EPSILON) moment_sum=REAL8m_EPSILON;  // Safety

  // Set up neighbor structure in m and m0, m1, m2
  int new_ngbrcnt;
  int maxmagngbrs=MagElt::GetMaxNgbrcnt();
  MagEltLink* Neighbors = new MagEltLink[maxmagngbrs];

  if(maxmagngbrs!=4 && maxmagngbrs!=8)
   PlainError(1,"Error in Grid2D::Grid2D(MIF &): "
	    "Current code only supports 4 and 8 neighborhoods");

  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m,Neighbors,i,k);
    m[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m0,Neighbors,i,k);
    m0[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m1,Neighbors,i,k);
    m1[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
    new_ngbrcnt=FindNeighbors(maxmagngbrs,m2,Neighbors,i,k);
    m2[i][k].SetupNeighbors(new_ngbrcnt,Neighbors);
  }
  delete[] Neighbors;
}

void Grid2D::Deallocate(void)
{
  int i;
  if(ExternalDemag->calc!=NULL) (*ExternalDemag->destroy)();
  ThreeVector::Release2DArray(Nx,h);
  ThreeVector::Release2DArray(Nx,h0);
  ThreeVector::Release2DArray(Nx,h1);
  ThreeVector::Release2DArray(Nx,anisa);
  ThreeVector::Release2DArray(Nx,anisb);
  ThreeVector::Release2DArray(Nx,anisc);
  if(DoDemag) ThreeVector::Release2DArray(Nx,hdemag);
  ThreeVector::Release2DArray(Nx,torque1);
  ThreeVector::Release2DArray(Nx,torque0);
  ThreeVector::Release2DArray(Nx,torque);
  m.Deallocate(); m0.Deallocate(); m1.Deallocate(); m2.Deallocate();
  for(i=0;i<Nx;i++) {
    if(ExternalDemag->calc==NULL && DoDemag)
      { delete[] ADemagCoef[i]; delete[] CDemagCoef[i]; }
  }
  if(ExternalDemag->calc==NULL && DoDemag)
    { delete[] ADemagCoef; delete[] CDemagCoef; }
  if(dfparams!=NULL) { delete dfparams[0]; delete dfparams; }
}

Grid2D::~Grid2D(void)
{
#ifdef STOPWATCH_ON
  PrintUsageTimes(stdout);
#endif
  delete hApplied;
  Deallocate();
}

void Grid2D::PrintUsageTimes(FILE *fptr)
{
#ifdef STOPWATCH_ON
  Oc_TimeVal cpu,wall;
  fprintf(fptr,"# Timing (seconds)     CPU       Elapsed\n");
  sw_exch.GetTimes(cpu,wall);
  fprintf(fptr,"#  Exchange    %12.3f %12.3f\n",
	  double(cpu),double(wall));
  sw_anis.GetTimes(cpu,wall);
  fprintf(fptr,"#  Anisotropy  %12.3f %12.3f\n",
	  double(cpu),double(wall));
  sw_demag.GetTimes(cpu,wall);
  fprintf(fptr,"#  Demag       %12.3f %12.3f\n",
	  double(cpu),double(wall));
  sw_solve_total.GetTimes(cpu,wall);
  fprintf(fptr,"#  Solve Total %12.3f %12.3f\n",
	  double(cpu),double(wall));
  sw_proc_total.GetTimes(cpu,wall);
  fprintf(fptr,"#  Proc  Total %12.3f %12.3f\n",
	  double(cpu),double(wall));
  fprintf(fptr,"Total number of field updates: %d\n",GethUpdateCount());
  fflush(fptr);
#endif // STOPWATCH_ON
}

void Grid2D::Perturb(double max_mag)
{ // Randomly perturb all m by <magnitude
  for(int i=0;i<Nx;i++) for(int k=0;k<Nz;k++) {
    m[i][k].Perturb(max_mag);
  }
  hValid=torqueValid=energyValid=0;
}

// The next two functions (ADemag() & CDemag()) access the demagcoef
// matrices using symmetries to allow negative values for i and k.
double Grid2D::ADemag(int i,int k)
{
#ifdef OLD_CODE
  return ADemagCoef[abs(i)][abs(k)];
#else
  return ADemagCoef[(i>=0?i:-i)][(k>=0?k:-k)];
#endif
}

double Grid2D::CDemag(int i,int k)
{
#ifdef OLD_CODE
  if(i==0 || k==0) return 0.0;
  double dtemp=CDemagCoef[abs(i)][abs(k)];
  if((i>0 && k>0) || (i<0 && k<0)) return  dtemp;
  else                             return -dtemp;
#else
  int sign=1;
  if(i<0) { i*=-1; sign=-1; }
  if(k<0) { k*=-1; sign*=-1; }
  return  sign*CDemagCoef[i][k];
#endif
}

/* Calculate the demag field at (i,k) including the self-field */
ThreeVector& Grid2D::CalculateDemag(int i,int k)
{
  static double hdarr[3];
  static ThreeVector hd(hdarr);
  if(!DoDemag) return hd;

  hd[0]=hd[1]=hd[2]=0.0;

  double A,C,mx,mz;
#ifndef UNIT_MAG
  double mom;
#endif
  int i2,k2;
  for(i2=0;i2<Nx;i2++) for(k2=0;k2<Nz;k2++) {
    A=ADemag(i2-i,k2-k); C=CDemag(i2-i,k2-k);
    mx=m[i2][k2][0];     mz=m[i2][k2][2];
#ifndef UNIT_MAG
    hd[0]+=(mom=m[i2][k2].GetMoment())*(A*mx+C*mz);
    hd[2]+=mom*(C*mx-A*mz);
#else
    hd[0]+=A*mx+C*mz;    hd[2]+=C*mx-A*mz;
#endif
  }
#ifndef UNIT_MAG
  mom=m[i][k].GetMoment();
  hd[0]+=mom*SelfDemag*m[i][k][0]; // Self-demag
  hd[2]+=mom*SelfDemag*m[i][k][2];
#else
  hd[0]+=SelfDemag*m[i][k][0]; // Self-demag
  hd[2]+=SelfDemag*m[i][k][2];
#endif
  return hd;
} 

/* Calculate the energy for the entire mesh */
void Grid2D::CalculateEnergy(void)
{
  int i,k;
  double exchcol,aniscol,demagcol,zeemancol;
  MagElt* mptr;
  double happmag,happarr[VECDIM]; ThreeVector happ(happarr);

  if(!hValid) hUpdate();

  exch_energy=anis_energy=demag_energy=zeeman_energy=0.0;
  exchcol=aniscol=demagcol=zeemancol=0.0;  // Just to shut up compiler

  for(i=0;i<Nx;i++) {
    exchcol=aniscol=demagcol=zeemancol=0.0;
    STOPWATCH_START(sw_exch);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++)
      exchcol+=mptr->CalculateExchangeEnergy();
    STOPWATCH_STOP(sw_exch);
    STOPWATCH_START(sw_anis);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++)
      aniscol+=(mptr->*MagElt::AnisotropyEnergy)();
    STOPWATCH_STOP(sw_anis);
    STOPWATCH_START(sw_demag);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++) {
      if(DoDemag) {
#ifndef UNIT_MAG
	demagcol+= mptr->GetMoment() * ((*mptr)*hdemag[i][k]);
	           // *(-1); see note (**) below
#else
	demagcol+=(m[i][k]*hdemag[i][k]);  // *(-1); see note (**) below
#endif // UNIT_MAG
      }
    }
    STOPWATCH_STOP(sw_demag);
    for(k=0,mptr=m[i];k<Nz;k++,mptr++) {
      hApplied->GetLocalhField(i,k,happ);
      ConvertXyzToXzy(happ);  // Coords change
      happmag=hApplied->GetLocalhFieldMag(i,k);
#ifndef UNIT_MAG
      zeemancol+=mptr->GetMoment()*(happmag-((*mptr)*happ));
#else
      zeemancol+=happmag-((*mptr)*happ);
#endif // UNIT_MAG
      // Zeeman is offset (by hFieldMag) so minimum energy is 0
    }
    // Accumulate by columns to improve precision
    exch_energy+=exchcol;      anis_energy+=aniscol;
    demag_energy+=demagcol;  zeeman_energy+=zeemancol;
  }
  // demag_energy gets divided by 2 because changing m also changes h.
  demag_energy/=2;

  // Normalize energies
  exch_energy/=moment_sum;
  anis_energy/=moment_sum;
  demag_energy/=(-moment_sum);    // Demag above off by minus sign  (**)
  zeeman_energy/=moment_sum;

  // Total energy density---
  energy = exch_energy+anis_energy+demag_energy+zeeman_energy;
  energyValid=1;
}

void Grid2D::InternalDemagCalc(void)
{
#ifdef CODE_CHECKS
  if(!DoDemag)
    PlainError(1,"Error in Grid2D::InternalDemagCalc: "
	     "This routine shouldn't be called because DoDemag is false");
#endif
  for(int i=0;i<Nx;i++) for(int k=0;k<Nz;k++)
    hdemag[i][k]=CalculateDemag(i,k);
}

void Grid2D::hUpdate(void)
{
#ifdef CODE_CHECKS
  if(hValid) PlainWarning("Warning in Grid2D::hUpdate: "
			  "Updating h even though hValid is TRUE.");
#endif // CODE_CHECKS
  int i,k;
  hUpdateCount++;

  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    hApplied->GetLocalhField(i,k,h[i][k]);
    ConvertXyzToXzy(h[i][k]);
  }
  STOPWATCH_START(sw_exch);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=m[i][k].CalculateExchange();
  STOPWATCH_STOP(sw_exch);
  STOPWATCH_START(sw_anis);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    h[i][k]+=(m[i][k].*MagElt::AnisotropyField)();
  STOPWATCH_STOP(sw_anis);
  if(DoDemag) {
    STOPWATCH_START(sw_demag);
    if(ExternalDemag->calc!=NULL) // Use external demag routine
      (*ExternalDemag->calc)(WriteGrid2Dm,FillGrid2Dh,m,hdemag);
    else // Use internal explicit (brute-force) demag calculation
      InternalDemagCalc();
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=hdemag[i][k];
    STOPWATCH_STOP(sw_demag);
  }

  hValid=1;
  torqueValid=energyValid=0;
}

void Grid2D::hFastUpdate(void)
{
  int i,k;

  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    hApplied->GetLocalhField(i,k,h[i][k]);
    ConvertXyzToXzy(h[i][k]);
  }
  STOPWATCH_START(sw_exch);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=m[i][k].CalculateExchange();
  STOPWATCH_STOP(sw_exch);
  STOPWATCH_START(sw_anis);
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    h[i][k]+=(m[i][k].*MagElt::AnisotropyField)();
  STOPWATCH_STOP(sw_anis);
  if(DoDemag) {
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) h[i][k]+=hdemag[i][k];
  }                                           // Use old demag
  torqueValid=energyValid=0;
}

void Grid2D::CalculateTorque(Array2D<MagElt> &m_in,
			       ThreeVector** &h_in,
			       ThreeVector** &t_out)
{ // Calculates torque from imports m_in & h_in, storing
  // result in t_out.
  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);
  int i,k;
  MagElt *mp;
  ThreeVector *hp;
  ThreeVector *tp;

  for(i=0;i<Nx;i++) {
    hp=h_in[i]; tp=t_out[i]; mp=m_in[i];
    for(k=0;k<Nz;k++,hp++,tp++,mp++) {
      vtemp= *mp;
      tp->FastAdd(*hp,-(vtemp*(*hp)),vtemp);
#ifdef PRECESSION
      vtemp^=(*hp);                 // Precession term
      tp->Accum(PRECESSION_RATIO,vtemp);
#endif // PRECESSION

#ifndef UNIT_MAG
      (*tp) *= mp->GetMoment();
#endif
    }
  }
}

double Grid2D::GetMaxTorque(void)
{
  int i,k;
  double dtemp,maxtorque=0.0;
  if(!hValid) hUpdate();
  if(!torqueValid) { CalculateTorque(m,h,torque); torqueValid=1; }
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    if((dtemp=torque[i][k].MagSq())>maxtorque) maxtorque=dtemp;
  }
  return sqrt(maxtorque);
}

void Grid2D::StepEuler(double maxtorque,double &next_stepsize)
{ // Takes one Euler step, with stepsize control based on energy.  Assumes
  // m, h, torque & energy are current coming in, and leaves them current
  // going out.  Saves entry values of m, h, and torque in m0, h0, and
  // torque0.
  //   The attempted step size is determined by the value of *this->StepSize
  // on entry.  StepSize may be adjusted by this routine; the value of
  // StepSize on exit is the step size actually used.  The export
  // next_stepsize should be used by the calling routine (on the next pass)
  // to set *this->StepSize _before_ calling this routine.

  const double CutRatio=0.5;        // Determined by trial & error
  const double IncreaseRatio=1.1;   //             ""
  // NOTE: I've tried more sophisticated stepsize correction, but this
  //       simple scheme seems to work as well as anything else.
  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);
  ThreeVector** htemp; //  Temp. pointer used for swapping h and torque
  double orig_energy,orig_stepsize=StepSize;
  double predicted_change,actual_change;  // Change in energy
  double error;
  int i,k;

  // Save current m, h and energy
  m.Swap(m0);
  htemp=h; h=h0; h0=htemp;
  orig_energy=energy;

  while(1) {
    hValid=0; predicted_change=0.0;
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      vtemp=m0[i][k];
      m[i][k].FastAdd(1.-OC_SQ(StepSize)*torque[i][k].MagSq()/2.,vtemp,
		      StepSize,torque[i][k]);
      m[i][k].Scale(1);
      vtemp-=m[i][k];
      predicted_change+=(vtemp*h0[i][k])*m[i][k].GetMoment();
      // Should the RHS by multiplied by m[i][k].GetMoment() ??? /**/
      // If I leave it off, I get occasional "Step too small" errors.
      // -mjd, 5-Dec-1996.
    }
    predicted_change*=-1; // Torque is -dE/dM
    hUpdate();  // m, h now hold new position
    CalculateEnergy();
    actual_change=energy-orig_energy;
    error=fabs(actual_change-predicted_change);
    if(actual_change<=0. && error<(AllowedSpinError*Nx)*Nz) break; // Good step
    StepSize*=CutRatio; // Bad step
    if((StepSize*maxtorque)<REAL8m_EPSILON) {  // Safety
      // Temporarily disable the following warning message until  /**/
      // we get a non-blocking log channel setup (26-Feb-1998).   /**/
      // PlainWarning("Warning in Grid2D::Euler: "
      //            "Step (%g) too small (maxtorque=%g)",StepSize,maxtorque);
      StepSize=0; next_stepsize=orig_stepsize;
      m.Swap(m0); htemp=h; h=h0; h0=htemp; // Undo changes
      energy=orig_energy;
      return;
    }
  }
  next_stepsize=StepSize*(1.+(IncreaseRatio-1.)*(1.-error));

  htemp=torque; torque=torque0; torque0=htemp; // Save old torque and
  CalculateTorque(m,h,torque);                 // calculate new torque
  torqueValid=1;
}

void Grid2D::StepPredict2(double maxtorque,double &next_stepsize)
{ // Takes one step using a 2nd order predictor, with stepsize
  // control.  Assumes m, h, torque & energy are current coming in, and
  // leaves them current going out.  Saves entry values of m, h, and
  // torque in m0, h0, and torque0. m1, h1, and torque1 are used for
  // temporary workspace.
  //   The attempted step size is determined by the value of
  // *this->StepSize on entry.  StepSize may be adjusted by this routine;
  // the value of StepSize on exit is the step size actually used.
  // The export next_stepsize should be used by the calling routine (on
  // the next pass) to set *this->StepSize _before_ calling this routine.
  //   If ODEIterCount<1, this routine just calls StepEuler.  Otherwise,
  // it assumes m0, torque0 and StepSize0 have valid values from the
  // previous step.  The controlling program *must* call ResetODE() to
  // reset ODEIterCount on each new integration segment (i.e., when
  // m0, torque0 or StepSize0 are invalid), or when a ODE history
  // restart is desired.

  const double CutRatio=0.5;        // Determined by trial & error
  const double IncreaseRatio=1.1;   //             ""

  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);

  if(ODEIterCount<1) { // Insufficient past history to do multistep
    StepEuler(maxtorque,next_stepsize);
    return;
  }

  ThreeVector** htemp; //  Temp. pointer used for swapping h and torque
  int i,k;
  double actual_energy_change;

  double orig_stepsize=StepSize;
  double orig_energy=energy;
  double relstep,tbasecoef,t0coef,tcoef,t1coef;

  while(1) {
    relstep=StepSize/StepSize0;

    // First estimate m1 & h1 in order to estimate torque1
    t0coef= -0.5*relstep*StepSize; tcoef=StepSize*(1+0.5*relstep);
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      //      h1[i][k].FastAdd((1+relstep),h[i][k],-relstep,h0[i][k]);
#ifdef OLDCODE
      m1[i][k].FastAdd(1.-OC_SQ(StepSize)*torque[i][k].MagSq()/2.,m[i][k],
      		       StepSize,torque[i][k]); // Includes correction from
                                               // |m1[i][k]| \in S^2
#else
//      m1[i][k].FastAdd(m[i][k],tcoef,torque[i][k]);
//      m1[i][k].Accum(t0coef,torque0[i][k]);
      vtemp.FastAdd(t0coef,torque0[i][k],tcoef,torque[i][k]);
      m1[i][k].FastAdd(vtemp,1.-vtemp.MagSq()/2.,m[i][k]);
      // Includes correction due to restriction that |m1[i][k]| \in S^2
#endif
      m1[i][k].Scale(1);
    }
    m1.Swap(m);
    htemp=h; h=h1; h1=htemp;
    hValid=0; hUpdate(); // Estimate h1 from m1 & hdemag_0
    m1.Swap(m);
    htemp=h; h=h1; h1=htemp;
    CalculateTorque(m1,h1,torque1);

    // Now calculate m1 using m, torque0, torque, and torque1
    tbasecoef=StepSize/(6.*(relstep+1));
    t0coef= -relstep*relstep;
    tcoef=  (relstep+4)*relstep+3;
    t1coef= 2*relstep+3;
    t0coef*=tbasecoef; tcoef*=tbasecoef; t1coef*=tbasecoef;
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      m1[i][k].FastAdd(t0coef,torque0[i][k],tcoef,torque[i][k]);
      m1[i][k].Accum(t1coef,torque1[i][k]);
      m1[i][k]+=m[i][k];
      m1[i][k].Scale(1);
    }
    hValid=0;
    m1.Swap(m); // Swap new m with m on entry
    htemp=torque; torque=torque1; torque1=htemp;
    htemp=h; h=h1; h1=htemp; // Save old field
    hUpdate();  // m, h now hold new position
    CalculateEnergy();
    CalculateTorque(m,h,torque);  // Calculate "real" new torque
    torqueValid=1;

    actual_energy_change=energy-orig_energy;
    if(actual_energy_change<0) break; // Good step
    // Else, bad step
    m1.Swap(m); htemp=h; h=h1; h1=htemp; // Undo changes
    htemp=torque; torque=torque1; torque1=htemp;
    energy=orig_energy;
    StepSize*=CutRatio;
    if((StepSize*maxtorque)<REAL8m_EPSILON) {  // Precision check
      // Temporarily disable the following warning message until  /**/
      // we get a non-blocking log channel setup (26-Feb-1998).   /**/
      // PlainWarning("Warning in Grid2D::StepPredict2: "
      //      "Step (%g) too small (maxtorque=%g)",StepSize,maxtorque);
      StepSize=0; next_stepsize=orig_stepsize;
      return;
    }
  }
  // Save entry values of m, h and torque
  m1.Swap(m0); htemp=h0; h0=h1; h1=htemp;
  htemp=torque0; torque0=torque1; torque1=htemp;

  next_stepsize=StepSize*(1.+(IncreaseRatio-1.)*1.0);
}

void Grid2D::IncRungeKutta4(Array2D<MagElt> &m_in,
			    ThreeVector** &h_in,
			    double step,
			    Array2D<MagElt> &m_out)
{ // Takes one 4th order Runge-Kutta step, w/o stepsize control.
  // Used as base stepper for Grid2D::StepRungeKutta4()
  // m_in and h_in are imports; results stored in m_out
  // Overwrites m, h, & torque (used for temporary storage)
  // NOTE: m_out *must* be distinct from m_in, and both must be distinct
  //       from m;  Also, h_in must be distinct from h.
  int i,k;
  
  CalculateTorque(m_in,h_in,torque); // Torque at starting location

  // Test step 1
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].FastAdd(m_in[i][k],step/6.0,torque[i][k]);
    m[i][k].FastAdd(m_in[i][k],step/2.0,torque[i][k]);
    m[i][k].Scale(1);  // Might want to delay rescaling, but if we do,
                       // will hUpdate() & CalculateTorque() work properly?
  }
#ifdef RUNGE_KUTTA_FAST
  hFastUpdate();  // m, h now hold first test location
#else
  hUpdate();      // m, h now hold first test location
#endif

  // Test step 2
  CalculateTorque(m,h,torque);
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].Accum(step/3.0,torque[i][k]);
    m[i][k].FastAdd(m_in[i][k],step/2.0,torque[i][k]);
    m[i][k].Scale(1);  // Ditto rescaling comment.
  }
#ifdef RUNGE_KUTTA_FAST
  hFastUpdate();  // m, h now hold second test location
#else
  hUpdate();      // m, h now hold second test location
#endif

  // Test step 3
  CalculateTorque(m,h,torque);
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].Accum(step/3.0,torque[i][k]);
    m[i][k].FastAdd(m_in[i][k],step,torque[i][k]);
    m[i][k].Scale(1);  // Ditto rescaling comment.
  }
#ifdef RUNGE_KUTTA_FAST
  hFastUpdate();  // m, h now hold third test location
#else
  hUpdate();      // m, h now hold third test location
#endif

  // Test step 4
  CalculateTorque(m,h,torque);
  hValid=0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    m_out[i][k].Accum(step/6.0,torque[i][k]);
    m_out[i][k].Scale(1);
  }
}

void Grid2D::StepRungeKutta4(double maxtorque,
			     double &next_stepsize,int forcestep)
{ // Takes one 4th order Runge-Kutta step, with adaptive stepsize
  // control.  Assumes m, h, torque & energy are current coming in, and
  // leaves them current going out.
  // The attempted step size is determined by the value of *this->StepSize
  // on entry.  StepSize may be adjusted by this routine; the value of
  // StepSize on exit is the step size actually used.  The export
  // next_stepsize should be used by the calling routine (on the next pass)
  // to set *this->StepSize _before_ calling this routine.
  //  Note: If the "forcestep" import is 1, then the smallest allowed step
  // is accepted, even if the error is above tolerance.  This really
  // shouldn't be used... ;^)

  const double SafetyFactor=0.8;  // Set by Trial & Error
  const double HeadRoom=0.9;      // Ditto
  static double vtemparr[3];  static ThreeVector vtemp(vtemparr);
  ThreeVector** htemp; //  Temp. pointer used for swapping h
  double halfstep,fullstep;
  double oldenergy,error,dtemp,orig_stepsize=StepSize;
  int i,k;

  // Save current m, h and energy
  m.Swap(m0); htemp=h; h=h0; h0=htemp;
  oldenergy=energy;

  // Do one valid step
  next_stepsize=StepSize;
  while(1) {
    if((next_stepsize*maxtorque)<REAL8m_EPSILON) { // Precision check
      if(forcestep) {       
	PlainWarning("Grid2D::StepRungeKutta4: "
		     "Step (%g) too small (maxtorque=%g);"
		     " continuing anyway...",next_stepsize,maxtorque);
      } else {
	PlainWarning("Grid2D::StepRungeKutta4: "
		     "Step (%g) too small (maxtorque=%g)",
		     next_stepsize,maxtorque);
	StepSize=0; next_stepsize=orig_stepsize;
	m.Swap(m0); htemp=h; h=h0; h0=htemp; energy=oldenergy; // Undo changes
	return;
      }
    }
    halfstep=(fullstep=next_stepsize)/2.0;
    // Do first half step, saving result in m1
    IncRungeKutta4(m0,h0,halfstep,m1);
    // Prepare for second half step
    m1.Swap(m); hValid=0; hUpdate();
    m.Swap(m1); htemp=h; h=h1; h1=htemp;
    // Do second half step, saving result in m2
    IncRungeKutta4(m1,h1,halfstep,m2);

    // Do full step, saving result in m1
    IncRungeKutta4(m0,h0,fullstep,m1);
    StepSize=fullstep;

    // Determine error estimate
    error=0; hValid=0;
    for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
      vtemp.FastAdd(m2[i][k],-1.0,m1[i][k]);
      if((dtemp=vtemp.MagSq())>error) error=dtemp;
#ifdef OLD_CODE
      m[i][k].FastAdd(m2[i][k],1./15.,vtemp);  // <== Here is final m!
#endif  // On 1st standard problem, straight m2 converges in fewer steps.
    }
#ifndef OLD_CODE
    m.Swap(m2);   // <== Here is final m!
#endif
    error=OC_MAX(sqrt(error),AllowedSpinError/1024.);
    // The "1024." above is arbitrary, used to insure error>0.
    // This also keeps next_stepsize from increasing by more than
    // a factor of 4 in one step.

    // Adjust step size based on error estimate & energy, and
    // terminate "one valid step" loop if appropriate.
    next_stepsize=fullstep*pow(AllowedSpinError/error,0.2)*SafetyFactor;
    if(forcestep && (next_stepsize*maxtorque)<REAL8m_EPSILON) {
      // nextstep too small, but forcestep is requested.
      hUpdate(); CalculateEnergy();
      break;
    }
    if(error>AllowedSpinError) continue; // Bad step; cut and retry
    hUpdate(); CalculateEnergy(); // Else check that energy is decreasing
    if(energy<oldenergy) break;   // Good step!
    next_stepsize=fullstep/2.0; // Bad step;  Cut and retry
  }
  next_stepsize*=HeadRoom;
  CalculateTorque(m,h,torque); torqueValid=1;
}

double Grid2D::StepODE(int& errorcode)
{ // Iterates one step, updating m, h, and energy.
  // Returns max torque value.
  // Will update h as needed.  Will update energy & torque only if
  // hValid is FALSE or ODEIterCount<1.  ***THIS SHOULD BE CONSIDERED***
  //                                     ***   A BUG\b\b\bFEATURE!   ***
  // errorcode>0 on error, =0 on success, <0 if successful with warnings
  STOPWATCH_START(sw_solve_total);
#define PERTURB_RETRIES 0  // Number of perturbation retries before
                          /// failing
  int perturb_count=0;
  double maxtorque;

  errorcode=0;
  StepSize0=StepSize; // Store size of previous step

  // Get max value of torque.  This will update h and
  // torque if necessary. 
  maxtorque=GetMaxTorque();

  // Update energy, if necessary.
  if(!energyValid) CalculateEnergy();

  // Take one step
  while(1) {
    if(NextStepSize>0) StepSize=NextStepSize;
    // Check stepsize bound
    if(StepSize*maxtorque>MaxTorqueStep) {
      StepSize=MaxTorqueStep/maxtorque;
    }
#ifdef EULER_ODE               // 1st order Euler
    StepEuler(maxtorque,NextStepSize);
#elif defined(RUNGE_KUTTA_ODE) // 4th order Runge-Kutta
    StepRungeKutta4(maxtorque,NextStepSize);
#elif defined(PREDICTOR_2_ODE)
    StepPredict2(maxtorque,NextStepSize);
#else
#error Unimplemented ODE solver request
#endif

    if(StepSize>0) break;  // Valid step found

#ifndef RUNGE_KUTTA_ODE
    // Couldn't find valid step.  Try a 4th order Runge-Kutta step
    ResetODE();
    StepSize=MaxTorqueStep/maxtorque;
    StepRungeKutta4(maxtorque,NextStepSize,1);
    if(StepSize>0.) break;  // Step accepted
#endif

    if(perturb_count>=PERTURB_RETRIES) break; // Give up

    // Otherwise, no valid step found.  Perturb and retry
    MessageLocker::Append("Warning in Grid2D::StepODE: "
			  "Unable to find a valid step,"
			  "  Will jiggle values and retry.\n");
    perturb_count++;
    Perturb(PERTURBATION_SIZE); hUpdate();  CalculateEnergy();
    CalculateTorque(m,h,torque); torqueValid=1;
    maxtorque=GetMaxTorque();
    if(NextStepSize*maxtorque>MaxTorqueStep)
      NextStepSize=MaxTorqueStep/maxtorque;
  }

  // Adjust NextStepSize as required by MaxTorqueStep
  maxtorque=GetMaxTorque();
  if(NextStepSize*maxtorque>MaxTorqueStep) {
    NextStepSize=MaxTorqueStep/maxtorque;
  }

  // Check for errors, and increment ODEIterCount if applicable
  if(StepSize<=0) {
    errorcode=1;
    MessageLocker::Append("ERROR in Grid2D::StepODE: "
			  "Unable to find a valid step (maxtorque=%g),"
			  "  Giving up.\n",maxtorque);
    PlainWarning("ERROR in Grid2D::StepODE: "
		 "Unable to find a valid step (maxtorque=%g),"
		 "  Giving up.\n",maxtorque);
  } else {
    ODEIterCount++;
    errorcode = -perturb_count;
  }

  STOPWATCH_STOP(sw_solve_total);
  return maxtorque;
#undef PERTURB_RETRIES
}

void Grid2D::GetNearestGridNode(double x,double y,int &i,int &j)
{
#ifdef CODE_CHECKS
  if(cellsize<REAL8m_EPSILON)
    PlainError(1,"Error in Grid2D::GetNearestGridNode: "
	     "Cellsize (=%g) is too small or not initialized",cellsize);
#endif // CODE_CHECKS
  i=int(ceil(x/cellsize));  j=int(ceil(y/cellsize));
  if(i<0) i=0; else if(i>=Nx) i=Nx-1;
  if(j<0) j=0; else if(j>=Nz) j=Nz-1;
}

void Grid2D::GetRealCoords(int i,int j,
			   double &xout,double &yout,double & zout)
{ // Node points are in cell centers.  Also note change of
  // coordinate systems, from xz+y to xy+z
  if(Nx<1) xout=0.0; else xout=xmax*double(i)/double(Nx);
  if(Nz<1) yout=0.0; else yout=zmax*double(j)/double(Nz);
  xout+=cellsize/2.; yout+=cellsize/2.;
  zout=0.;
}

double Grid2D::GetMaxNeighborAngle(void)
{ // Returns maximum angle (in radians) between
  // neighboring spin directions in the grid.
  int i,k;
  double dot,mindot;
  mindot=1.0;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) {
    // Note: Spins with zero-moment *should* have
    //  no neighbors, and therefore return dot=1.0
    dot=m[i][k].GetMinNeighborDot();
    if(dot<mindot) mindot=dot;
  }
  if(fabs(mindot)>1) {
    fprintf(stderr,"ERROR: Mindot=%15.12f\n",mindot);
    PlainError(1,"Error in Grid2D::GetMaxNeighborAngle: %s",ErrProgramming);
  }
  return acos(mindot);
}

Vec3D Grid2D::GetAveMag()
{
  int i,k;
  Vec3D AveMag(0.,0.,0.);
#ifndef UNIT_MAG
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    AveMag.Accum(m[i][k].GetMoment(),m[i][k]);
#else
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++) AveMag+=m[i][k];
#endif
  AveMag*=(1.0/moment_sum);
  AveMag.ConvertXzyToXyz();
  return AveMag;
}

void Grid2D::GetAverageMagnetization(ThreeVector& AveMag)
{
  Convert(AveMag,GetAveMag());  // Note that the conversion
  /// from xz+y coords to xy+z coords is done inside GetAveMag().
}

#ifdef CONTRIB_SPT001
double Grid2D::GetResistance(void)
{ // Returns 'resistance' for the following assumptions 1) We are
  // examining the AMR effect 2) It is a THIN film being examined 3) The
  // individual 'cells' are in a SINGLE domain state i.e. the
  // magnetization accross a 'cell' is uniform - I believe this is the
  // assuption the solver works on (correct ?).
  //   For this special case we can say the resistance of a 'cell' is
  // equal to: R0 + (deltaR*((cos theta)^2))) Where R0 is the base
  // resistance, deltaR is the maximum change in resistance (typically
  // about %2) and theta is the angle between the 'cell' magnetization
  // and 'current'.
  //   Version 1 of this function merely returns <(cos theta)^2> of the
  // magnetizations and the 'current'. Current is currently 'fixed' as
  // (1,0,0) in xyz co-ords (need to add it to a MIF input file). The
  // next stage is to have R0 and deltaR defined as material parameters,
  // and then remove the definition of the values from this function to
  // handle 'real' resistance characteristics.
  //   Problems: Doesn't handle masks yet. Masks have a zero magnetization
  // vector. All it would do is to 'dilute' the average of the
  // resistance.

  int i,k;
  double R1,Rt,magCheck,currCheck,R0,deltaR; // R0 and deltaR are the
  /// resistance properties of the material and should be incorporated
  /// into the MIF description.  One question is will the MIF file define
  /// deltaR or deltaR/R0 ?

  Vec3D Mag(0.,0.,0.);
  Vec3D Current(1.,0.,0.); // Currently a vector in xyz co-ords (MIF input
                          /// format). Doesn't have to be a unit vector.

  Current.ConvertXyzToXzy(); // Keep everything INSIDE Grid2D as xzy
                            /// co-ords.

  R1=Rt=(double)0.0;
  R0=(double)0.0;     // Fix R0 and deltaR such that the function
                      // currently returns just <(cos theta)^2>,
  deltaR=(double)1.0; // the same as before until these values can be
                      // incorporated into material parameters.

  // deltaR*=R0;     // This conversion used to scale from MR co-efficient
                     // just once, if that system is used.

  currCheck=Current.MagSq();					
  currCheck=(!currCheck)?0:currCheck=sqrt(currCheck);
  /// Does Sqrt blow up for sqrt(0) ? Let's not find out.
  /// This has been moved to out of the loop for efficiency.
  /// It implies the current is uniform accross the grid.
  for(i=0;i<Nx;i++) {
    for(k=0;k<Nz;k++) {
      Mag=m[i][k];

      magCheck=m[i][k].GetMoment();
      /// Check for modulus of zero on the magnetization.

      // Make sure we dont break the oommf by trying to divide by zero
      if( (magCheck != 0) && (currCheck != 0) ) {
	R1=(Mag*Current);	// Dot the two vectors together.
	R1/=currCheck;
 	/// Divide by the modulus' of the Current:(Mag is a unit vec).
	R1*=R1;			// Square the result -> ((Cos Theta)^2).
      } else {
	// One of the vectors is zero - How to deal with this undefined
	// behaviour ?
	R1 = 0;			// Just set resistance to 0 currently.
      }
      Rt+=(R1/(Nx*Nz));	 // Add a fraction to the total (equiv to summing
                         // all and dividing by number summed).
    }   
  }

  return (R0+deltaR*Rt); // Maximum resistance when Mag and Current are
  /// parallel, min when perpendicular when the MR co-efficient is positive.
}
#endif // CONTRIB_SPT001

void Grid2D::GetMag(double** arr)
{ // Fills arr with magnetizations.  The calling routine must
  // make sure arr is big enough (cf. Grid2D::GetDimens()).
  int i,k;
  for(i=0;i<Nx;i++) for(k=0;k<Nz;k++)
    m[i][k].DumpMag(arr[i]+k*VECDIM);
}

void Grid2D::Dump(CHAR *filename,CHAR *note)
{
  int i,k;
  FILE *fptr;
  double Bapparr[3]; ThreeVector Bapp(Bapparr);
  hApplied->GetLocalBField(0,0,Bapp);

  if((fptr=Nb_FOpen(filename,"wb"))==NULL) {
    fprintf(stderr,"FATAL ERROR: Unable to open output file %s\n",
	    filename);
    exit(1);
  }

  double maxtorque=GetMaxTorque();
  double e_exch,e_anis,e_demag,e_zeeman,e_total;
  GetEnergies(e_exch,e_anis,e_demag,e_zeeman,e_total);

  if(note!=NULL) fprintf(fptr,"%s\n\n",note);
  fprintf(fptr,
	  "\nMicromagnetic grid parameters-------\n"
	  " Grid size                     %3dx%d\n"
	  " Demag field routine        demag = %s\n"     
	  " Saturation magnetization      Ms = %g (A/m)\n"
	  " Anisotropy constant           K1 = %g (J/m^3)\n"
	  " Exchange stiffness             A = %g (J/m)\n"
	  " Applied field               Bapp = <%g,%g,%g> (T)\n"
	  " Mesh cell edge length       mesh = %g (m)\n"
	  " Final step size         StepSize = %g\n\n",
	  Nx,Nz,ExternalDemag->name,Ms,K1,exchange_stiffness,
	  Bapp[0],Bapp[1],Bapp[2],
	  cellsize,StepSize);
  fprintf(fptr,"Micromagnetic modeling results-------\n");
  fprintf(fptr,
	  "Energy:  Total = %10.6f\n"
	  "          Exch = %10.6f\n"
	  "          Anis = %10.6f\n"
	  "         Demag = %10.6f\n"
	  "        Zeeman = %10.6f\n",
	  e_total,e_exch,e_anis,e_demag,e_zeeman);
  fprintf(fptr,"Max torque: %10.7g\n",maxtorque);
#ifndef UNIT_MAG
  double mom;
#endif
  for(i=0;i<Nx;i++) {
    fprintf(fptr,"\nRow %d:\n",i);
    for(k=0;k<Nz;k++) {
      // Print out moment and torques.  Note change of coords
      // from internal xz+y to xy+z.
#ifndef UNIT_MAG
      mom=m[i][k].GetMoment();
      fprintf(fptr,"m[%2d][%2d]=(%7.4f,%7.4f,%7.4f)"
	      "\tt[%2d][%2d]=(%10.7f,%10.7f,%10.7f)\n",
	      i,k,mom*m[i][k][0],mom*m[i][k][2],-mom*m[i][k][1],
	      i,k,torque[i][k][0],torque[i][k][2],-torque[i][k][1]);
#else
      fprintf(fptr,"m[%2d][%2d]=(%7.4f,%7.4f,%7.4f)"
	      "\tt[%2d][%2d]=(%10.7f,%10.7f,%10.7f)\n",
	      i,k,m[i][k][0],m[i][k][2],-m[i][k][1],
	      i,k,torque[i][k][0],torque[i][k][2],-torque[i][k][1]);
#endif // UNIT_MAG
    }
  }

  fprintf(fptr,"\n\n");
  fclose(fptr);
}

// Binary dump of (reduced) magnetization (VecFil format)
void Grid2D::BinDumpm(CHAR *filename,CHAR *note)
{
  int i;
  FileVector fv;

  fv.SetNote((BYTE *)note);

  if(fv.WriteHeader(filename,Nx,1,Nz,VECDIM)) {
    PlainWarning("Warning in Grid2D::BinDumpm: %s",ErrBadFile);
    return;
  }

  int k;
  double *temparr= new double[VECDIM*Nz];
  for(i=0;i<Nx;i++) {
    for(k=0;k<Nz;k++) m[i][k].DumpMag(temparr+k*VECDIM);
    if(fv.WriteVecs(Nz,temparr)) {
      PlainWarning("Warning in Grid2D::BinDumpm: %s",ErrBadFile);
      break;
    }
  }
  delete temparr;
  fv.Close();
}

// Binary dump of (reduced) field (VecFil format)
void Grid2D::BinDumph(CHAR *filename,CHAR *note)
{
  int i;
  FileVector fv;

  if(!hValid) { hUpdate(); CalculateEnergy(); }

  fv.SetNote((BYTE *)note);

  if(fv.WriteHeader(filename,Nx,1,Nz,VECDIM)) {
    PlainWarning("Warning in Grid2D::BinDumph: %s",ErrBadFile);
    return;
  }

  for(i=0;i<Nx;i++) {
    if(fv.WriteVecs(Nz,&(h[i][0][0]))) { // NOTE: &(h[i][0][0])     /**/
                                         // is a *BAD* thing        /**/
      PlainWarning("Warning in Grid2D::BinDumph: %s",ErrBadFile);
      break;
    }
  }
  fv.Close();
}

// Writes a 4-byte binary ovf file, of either m (if type=='m') or
// h (if type=='h').  The datastyle import should be one of "binary4",
// "binary8" or "text", with a NULL pointer or empty string implying
// the default value "binary4".  NB: the ovf format uses x & y in plane,
// as opposed to Grid2D which uses x & z in plane (both are righthand
// systems).  This routine does the necessary conversions.

void Grid2D::WriteOvf(CHAR type,const CHAR *_datastyle,
		      const CHAR *filename,const CHAR *realname,
		      const CHAR *note)
{
  if((type!='m' && type!='h') || filename==NULL) {
    PlainWarning("Warning in Grid2D::WriteOvf: %s",ErrBadParam);
  }

  Nb_DString datastyle("binary4");
  if(_datastyle!=NULL || _datastyle[0]!='\0') {
    datastyle.Dup(_datastyle);
  }
  datastyle.StripWhite(); datastyle.ToLower();
  if(strcmp(datastyle.GetStr(),"binary4")!=0 &&
     strcmp(datastyle.GetStr(),"binary8")!=0 &&
     strcmp(datastyle.GetStr(),"text")!=0) {
    PlainWarning("Warning in Grid2D::WriteOvf; Unknown data style: %s",
		 datastyle.GetStr());
  }

  FILE *fptr=Nb_FOpen(filename,"wb");
  if(fptr==NULL) {
    PlainWarning("Warning in GridDWriteOvf: Unable to open file %s"
		 " for writing",filename);
  }

  if(type=='h' && !hValid) {
    hUpdate(); CalculateEnergy();
  }

  if(realname==NULL) realname=filename;

  // Write ovf header /////////////////////////////////////////////////
  fprintf(fptr,
	  "# OOMMF: rectangular mesh v0.0a0\n"
	  "# Segment count: 1\n"
	  "# Begin: Segment\n"
	  "# Begin: Header\n");
  fprintf(fptr,
	  "# Title: %s\n",realname);
  if(note!=NULL && note[0]!='\0') {
    if(strchr(note,' ')!=NULL) {  // Spaces-only => no Desc:
      char *buf=new char[strlen(note)+1];
      strcpy(buf,note);
      char *cptr=strtok(buf,"\n");
      while(cptr!=NULL) {
	fprintf(fptr,"# Desc: %s\n",cptr);
	cptr=strtok(NULL,"\n");
      }
      delete[] buf;
    }
  } else {
    fprintf(fptr,"# Desc: Reduced %s\n",
	    (type=='m' ? "magnetization" : "field"));
  }

  // Note the coordinate system transformation: y -> -z
  fprintf(fptr,
	  "# meshunit: m\n"
	  "# xbase: %10g\n"
	  "# ybase: %10g\n"
	  "# zbase: %10g\n",cellsize/2.,cellsize/2.,ymax/2.);
  fprintf(fptr,
	  "# xstepsize: %10g\n"
	  "# ystepsize: %10g\n"
	  "# zstepsize: %10g\n",cellsize,cellsize,ymax);
  fprintf(fptr,
	  "# xnodes: %d\n"
	  "# ynodes: %d\n"
	  "# znodes: %d\n",Nx,Nz,1);
  fprintf(fptr,
	  "# xmin: 0.\n"
	  "# ymin: 0.\n"
	  "# zmin: 0.\n");
  fprintf(fptr,
	  "# xmax: %10g\n"
	  "# ymax: %10g\n"
	  "# zmax: %10g\n",xmax,zmax,ymax);
  fprintf(fptr,
	  "# valueunit: A/m\n"
	  "# valuemultiplier: %10g\n",Ms);
  if(type=='m') {
    fprintf(fptr,
	    "# ValueRangeMinMag: 1e-8\n"
	    "# ValueRangeMaxMag: 1.0\n");
  } else {
    // Determine largest field magnitude
    double max_mag=0,mag;
    for(int i=0;i<Nx;i++) for(int k=0;k<Nz;k++) {
      mag=h[i][k].MagSq();
      if(mag>max_mag) max_mag=mag;
    }
    max_mag=sqrt(max_mag);
    fprintf(fptr,
	    "# ValueRangeMinMag: %10g\n"
	    "# ValueRangeMaxMag: %10g\n",max_mag*1e-8,max_mag);
  }
  fprintf(fptr,
	  "# End: Header\n");

  // Dump data ////////////////////////////////////////////////////////
  if(strcmp(datastyle.GetStr(),"binary4")==0) {
    // 4-byte binary format //
    fprintf(fptr,
	    "#Begin: Data Binary 4\n");
    REAL4 checkvalue=1234567.;
# if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB
    flip4(&checkvalue);
# endif
    fwrite((char *)&checkvalue,sizeof(REAL4),1,fptr);
    union { REAL4 x; UCHAR c[4]; } outarr[3];
    if(sizeof(outarr[0])!=4) {
      PlainError(1,"Error in Grid2D::WriteOvf; "
		 "PROGRAMMING ERROR---REAL4 union %d bytes wide",
		 sizeof(outarr[0]));
    }
    if(type=='m') {
      // Write reduced magnetization
      double mtemp[3];
      for(int k=0;k<Nz;k++) for(int i=0;i<Nx;i++) {
	m[i][k].DumpMag(mtemp);  // Includes "relative moment" adjustment
	outarr[0].x=mtemp[0]; outarr[1].x=mtemp[2]; outarr[2].x= -mtemp[1];
	/// Note coordinate change.
#     if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB
	flip4(outarr[0].c); flip4(outarr[1].c); flip4(outarr[2].c);
#     endif // BYTEORDER
	fwrite((char *)outarr,sizeof(REAL4),3,fptr);
	/// The above fwrite ASSUMES outarr is packed.
      }
    } else {
      // Write reduced field
      for(int k=0;k<Nz;k++) for(int i=0;i<Nx;i++) {
	outarr[0].x=  h[i][k][0];
	outarr[1].x=  h[i][k][2];
	outarr[2].x= -h[i][k][1];
	/// Note coordinate change.
#     if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB
	flip4(outarr[0].c); flip4(outarr[1].c); flip4(outarr[2].c);
#     endif // BYTEORDER
	fwrite((char *)outarr,sizeof(REAL4),3,fptr);
	/// The above fwrite ASSUMES outarr is packed.
      }
    }
    fprintf(fptr,"\n"
	    "#End: Data Binary 4\n");

  } else if(strcmp(datastyle.GetStr(),"binary8")==0) {
    // 8-byte binary format //
    fprintf(fptr,
	    "#Begin: Data Binary 8\n");
    REAL8 checkvalue=123456789012345.;
# if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB
    flip8(&checkvalue);
# endif
    fwrite((char *)&checkvalue,sizeof(REAL8),1,fptr);
    union { REAL8 x; UCHAR c[8]; } outarr[3];
    if(sizeof(outarr[0])!=8) {
      PlainError(1,"Error in Grid2D::WriteOvf; "
		 "PROGRAMMING ERROR---REAL8 union %d bytes wide",
		 sizeof(outarr[0]));
    }
    if(type=='m') {
      // Write reduced magnetization
      double mtemp[3];
      for(int k=0;k<Nz;k++) for(int i=0;i<Nx;i++) {
	m[i][k].DumpMag(mtemp);  // Includes "relative moment" adjustment
	outarr[0].x=mtemp[0]; outarr[1].x=mtemp[2]; outarr[2].x= -mtemp[1];
	/// Note coordinate change.
#     if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB
	flip8(outarr[0].c); flip8(outarr[1].c); flip8(outarr[2].c);
#     endif // BYTEORDER
	fwrite((char *)outarr,sizeof(REAL8),3,fptr);
	/// The above fwrite ASSUMES outarr is packed.
      }
    } else {
      // Write reduced field
      for(int k=0;k<Nz;k++) for(int i=0;i<Nx;i++) {
	outarr[0].x=  h[i][k][0];
	outarr[1].x=  h[i][k][2];
	outarr[2].x= -h[i][k][1];
	/// Note coordinate change.
#     if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB
	flip8(outarr[0].c); flip8(outarr[1].c); flip8(outarr[2].c);
#     endif // BYTEORDER
	/// Note coordinate change.
	fwrite((char *)outarr,sizeof(REAL8),3,fptr);
	/// The above fwrite ASSUMES outarr is packed.
      }
    }
    fprintf(fptr,"\n"
	    "#End: Data Binary 8\n");

  } else if(strcmp(datastyle.GetStr(),"text")==0) {
    // Text data format //
    fprintf(fptr,
	    "#Begin: Data Text\n");
    if(type=='m') {
      // Write reduced magnetization
      double mtemp[3];
      for(int k=0;k<Nz;k++) for(int i=0;i<Nx;i++) {
	m[i][k].DumpMag(mtemp);  // Includes "relative moment" adjustment
	fprintf(fptr,"%10g %10g %10g\n",mtemp[0],mtemp[2],-mtemp[1]);
	/// Note coordinate change.
      }
    } else {
      // Write reduced field
      for(int k=0;k<Nz;k++) for(int i=0;i<Nx;i++) {
	fprintf(fptr,"%10g %10g %10g\n",h[i][k][0],h[i][k][2],-h[i][k][1]);
	/// Note coordinate change.
      }
    }
    fprintf(fptr,
	    "#End: Data Text\n");

  } else {
    // Safety: Should never get here!
    fclose(fptr);
    PlainError(1,"Error in Grid2D::WriteOvf; "
	       "PROGRAMMING ERROR---Bad filestyle: %s",datastyle.GetStr());
  }
  fprintf(fptr,"#End: Segment\n");
  fclose(fptr);
}

void WriteGrid2Dm(double** dat,void* m,int xdim,int ydim,int xyz)
{
  int i,k;
  int component=0,sign=1;
  switch(xyz) // Adjust for differences in component ordering
    {
    case 0: component=0;  sign= 1; break;
    case 1: component=2;  sign= 1; break;
    case 2: component=1;  sign=-1; break;  // Actually, this sign shouldn't
	    // matter because it gets reversed back in FillGrid2Dh()
    default: PlainError(1,"Error in WriteGrid2Dm: Illegal component request: %d",xyz);
	    break;
    }
#ifndef UNIT_MAG
  MagElt* mptr;
  for(i=0;i<xdim;i++) for(k=0,mptr=((MagElt**)m)[i];k<ydim;k++,mptr++)
    dat[i][k]=sign*(mptr->GetMoment())*((*mptr)[component]);
#else
  for(i=0;i<xdim;i++) for(k=0;k<ydim;k++)
    dat[i][k]=sign*((MagElt**)m)[i][k][component];
#endif

#ifdef EDGE_ADJUST
/**/ // Soften demag on edges
    for(k=0;k<ydim;k++) { 
      dat[0][k]*=EDGE_ADJUST; dat[xdim-1][k]*=EDGE_ADJUST;
    }
    for(i=1;i<xdim-1;i++) {
      dat[i][0]*=EDGE_ADJUST; dat[i][ydim-1]*=EDGE_ADJUST;
    }
#endif // EDGE_ADJUST
}

void FillGrid2Dh(double** dat,void* h,int xdim,int ydim,int xyz)
{
  int i,k;
  int component=0,sign=1;
  switch(xyz) // Adjust for differences in component ordering
    {
    case 0: component=0; sign= 1; break;
    case 1: component=2; sign= 1; break;
    case 2: component=1;  sign=-1; break;  // Actually, this sign shouldn't
	    // matter because it gets reversed back in FillGrid2Dh()
    default: PlainError(1,"Error in FillGrid2Dm: Illegal component request: %d",xyz);
	    break;
    }

#ifdef EDGE_ADJUST
/**/ // Soften demag on edges
    for(k=0;k<ydim;k++) { 
      dat[0][k]*=EDGE_ADJUST; dat[xdim-1][k]*=EDGE_ADJUST;
    }
    for(i=1;i<xdim-1;i++) {
      dat[i][0]*=EDGE_ADJUST; dat[i][ydim-1]*=EDGE_ADJUST;
    }
#endif // EDGE_ADJUST

  for(i=0;i<xdim;i++) for(k=0;k<ydim;k++)
    ((ThreeVector**)h)[i][k][component]=sign*dat[i][k];
}
