# FILE: procs.tcl
#
# A collection of Tcl procedures (not Oc_Classes) which are part of the
# Ow extension

########################################################################
# Routine that imports a color in any valid Tcl color format, and
# returns a string of the form #RRRRGGGGBBB representing the actual
# values that color will produce in the import window (which defaults
# to '.').
########################################################################
proc Ow_GetRGB { color {win .}} {
    foreach {r g b} [winfo rgb $win $color] {}
    return [format "#%04X%04X%04X" $r $g $b]
}

########################################################################
# Fixed point number verification routine.
########################################################################
# Returns 1 is $x is a number of the form d.d where d is a string
# of digits, -1 if $x is of the form -d.d, and zero otherwise.
# NOTE 1: Empty strings and a bare "." are allowed.  This is so
#   this function can be used to employ write restrictions in entry
#   widgets (its main role).
# NOTE 2: Whitespace is *not* allowed!
proc Ow_IsFixedPointEdit { x } {
    if {![regexp -- \
            {^(-?)[0-9]*\.?[0-9]*$} $x dummy sign]} {
	return 0
    }
    if {[string match "-" $sign]} {
	return -1
    }
    return 1
}

########################################################################
# Menu creation helper proc, Tk version independent.
########################################################################
proc Ow_MakeMenubar { rootwin menuwin args } {
    #    $args should be a list of triples.  Each triple is the name
    # (window path) of a submenu to make, the label to associate
    # with that submenu, and the index in the label of the letter
    # to underline for keyboard access.  If only one value is given,
    # it is assumed to be the label; the submenu path will be
    # constructed by appending the lowercase form of the label to
    # "${menuwin}.", and underline will be set to 0.  If only 2 values
    # are given, then if the first begins with a '.' it is assumed to
    # be the submenu path, label will be set from the second value,
    # underline will be set to 0.  If 2 values are given but the first
    # doesn't begin with a '.', then label and underline are set from
    # the input and submenu is constructed from the label as before.
    #    Submenus are created by this proc with '-tearoff 0'.  If you
    # want tearoff enabled, either create the submenu with '-tearoff 1'
    # before calling this proc, or (preferably) use 'configure -tearoff 1'
    # on the submenu after this proc returns.
    #
    # The return value is a list of the submenu pathnames, in order.
    #
    # Examples:
    #           Ow_MakeMenu . .menu File Options Help
    #           Ow_MakeMenu . .m {"Menu A" 5} {"Menu B" 5"} {"Menu C" 5}
    #           Ow_MakeMenu . .m {"A Menu"} {"B Menu"} {"C Menu"}
    #
    # Note that the last example requires 2 grouping sets, because a bare
    # "A Menu" or {A Menu} will be parsed as a two element list, and the
    # above rules will set underline to "Menu".
    #
    # Tk VERSION NOTE: The ".toplevel configure -menu" command first appears
    #   in Tcl/Tk version 8.0.  Prior to this menus had to be configured using
    #   a complicated arrangement of menus and menubuttons.  The Ow_MakeMenu
    #   proc detects which version of Tk is in use, and configures the menubar
    #   as appropriate.  Unfortunately, the older method requires an additional
    #   level of windows.  In this setting the import 'menuwin' is actually
    #   a frame, and the first argument in each of the $args triples is the
    #   winpath of a menubutton.  The actual submenus are children of the
    #   menubuttons, with extension .menu.  In all cases the return value is
    #   a list of _menus_ (never menubuttons).  Since the winpaths of the
    #   submenus will depend on the Tk version, it is *required* to setup the
    #   submenus based on the return value from this routine.  For example,
    #
    #        foreach {filemenu optionmenu helpmenu} \
    #                {[Ow_MakeMenu . .menu File Options Help]} {}
    #        $filemenu add command -label "Exit" -command { exit } -underline 1
    #        ...
    #
    #   will set filemenu to .menu.file under Tk 8.0, but .menu.file.menu under
    #   Tk 7.6.  Furthermore, .menu is a toplevel menu object in Tk 8.0, but a
    #   frame in Tk 7.6.  One may use the expression
    #
    #        if {[string match .menu [winfo toplevel .menu]]==0} { ... }
    #
    #   to distinguish the two cases.
    #

    global tk_version
    set new_style 1
    if {$tk_version<8.0} { set new_style 0}

    if {$new_style} {
        if {![winfo exists $menuwin]} { menu $menuwin }
        $rootwin configure -menu $menuwin
    } else {
        if {![winfo exists $menuwin]} {
            frame $menuwin -borderwidth 2 -relief raised
            pack $menuwin -side top -fill x -expand 0
        }
    }

    set submenu_list {}

    foreach i $args {
        set submenu {} ; set underline 0
        if {[llength $i]<2} {
            # Only label provided
            set label [lindex $i 0]
        } elseif {[llength $i]==2} {
            if {[string match ".*" [lindex $i 0]]} {
                # submenu and label provided
                set submenu [lindex $i 0]
                set label [lindex $i 1]
            } else {
                # label and underline provided
                set label [lindex $i 0]
                set underline [lindex $i 1]
            }
        } else {
            # All three fields provided
            set submenu [lindex $i 0]
            set label  [lindex $i 1]
            set underline [lindex $i 2]
        }
        if {[string match {} $submenu]} {
            set submenu "${menuwin}.[string tolower $label]"
        }
        if {$new_style} {
            if {![winfo exists $submenu]} { menu $submenu -tearoff 0}
            $menuwin add cascade -menu $submenu -label $label \
                    -underline $underline
            lappend submenu_list $submenu
        } else {
            set realmenu ${submenu}.menu
            if {![winfo exists $submenu]} {
                menubutton $submenu -text $label -menu $realmenu \
			-underline $underline
                if {[string compare $submenu ${menuwin}.help]==0} {
                    pack $submenu -side right ;# Pack "help" menu on the right.
                } else {
                    pack $submenu -side left
                }
            }
            if {![winfo exists $realmenu]} {
                menu $realmenu -tearoff 0
            }
            lappend submenu_list $realmenu
        }
    }
    return $submenu_list
}

#####################################################################
### Proc to position a child window relative to a parent.  The    ###
### imports xrat & yrat are the position of the upper lefthand    ###
### corner of the child relative to the ulh corner of the parent. ###
### Bug: if the parent geometry has negative position parameters, ###
### then the aforementioned corners will not be ulh.  Also, this  ###
### may be wm-specific.                                           ###
#####################################################################
proc Ow_PositionChild { child {parent .} {xrat .25} {yrat .25} } {
    set parentpos [wm geometry $parent]
    set junk ""
    regexp {^([0-9]*)x?([0-9]*)([-+][0-9]+)([-+][0-9]+)} \
            "$parentpos" junk parwidth parheight parposx parposy
    if { "$junk" != "" } then {
        if { "$parposx" >= 0 } then {
            set childposx [expr round($parposx + $xrat*$parwidth)]
        } else {
            set childposx [expr round($parposx - $xrat*$parwidth)]
        }
        # Note: Negative entries measure from lower & right window corner
        if { "$parposy" >= 0 } then {
            set childposy [expr round($parposy + $yrat*$parheight)]
        } else {
            set childposy [expr round($parposy - $yrat*$parheight)]
        }
        # We can write some fancier code (like centering) at a later date
        # -mjd 19-Jan-1997
        if { $childposx >= 0 } then { set childposx "+$childposx" }
        if { $childposy >= 0 } then { set childposy "+$childposy" }
        wm geometry $child "$childposx$childposy"
    }
}

###############################################
### Routine to associate icon with a window ###
### Under Windows this is a nop.            ###
###############################################
set ow(icon) [file join $ow(library) omficon.xbm]
set ow(iconmask) [file join $ow(library) omfmask.xbm]
proc Ow_SetIcon { win {iconname {}} {iconmaskname {}} } {
    global ow tcl_platform
    if {[string match windows $tcl_platform(platform)] \
            || ![winfo exists $win]} {
        return
    }
    set win [winfo toplevel $win]
    if { "$win" == "." } {
        set iconwin ".root:iconwindow"
    } else {
        set iconwin "$win:iconwindow"
    }

    if {[string match {} $iconname]} {
        set iconname $ow(icon)
    }
    if {[string match {} $iconmaskname]} {
        set iconmaskname $ow(iconmask)
    }

    if {[string match {} $iconname] || [string match {} $iconmaskname]} {
        return         ;# Give up
    }

    # Setup plain icon
    wm iconbitmap $win @$iconname
    wm iconmask $win @$iconmaskname

    # Setup fancier icon (not all wm's support this)
    if [winfo exists $iconwin] {
        destroy $iconwin
        update idletasks
    }
    if {![winfo exists $iconwin]} {
        toplevel $iconwin
        label $iconwin.label
        pack $iconwin.label
        $iconwin.label configure -bitmap @$iconname -bg yellow -fg black
        if {[catch {wm iconwindow $win $iconwin}] || \
                [string compare $iconwin [wm iconwindow $win]]!=0 || \
                [update idletasks ; \
                expr [winfo viewable $win] && [winfo viewable $iconwin]] || \
                [string compare "icon" [wm state $iconwin]]!=0 } {
            # wm iconwindow didn't take
            catch {wm iconwindow $win {}}
            destroy $iconwin
        } else {
            # wm iconwindow *did* take.  Setup auto-destruct, taking care
            # to ignore child window <Destroy> events
            bind $win <Destroy> "+
            if \[string compare $win %W\] continue
            # Added catch below to stop errors under RedHat Linux fvwm
            catch {destroy $iconwin}
            "
        }
    }
    # Here's an alternate way, by loading the image from a file:
    #    set imagetoken [image create photo -file oommficon.gif]
    #    toplevel $iconwin
    #    label $iconwin.label -image $imagetoken
    #    pack $iconwin.label
    #    wm iconwindow $win $iconwin
    #    unset imagetoken
}

######################################################
# Routines to push/pop watch cursor onto all currently
# existing toplevel windows.
#
set _watch_cursor_count 0
set _watch_cursor_safetyid {}
proc Ow_PushWatchCursor {} {
    global _watch_cursor_count _watch_cursor_safetyid
    if {$_watch_cursor_count<1} {
	# First watch cursor set event.
	. config -cursor watch
	foreach win [winfo children .] {
	    if {[string compare $win [winfo toplevel $win]]==0 && \
		    [string compare Menu [winfo class $win]]!=0} {
		$win config -cursor watch
		## Note: Setting the cursor on menu widgets on the
		##   winalp platform causes the enclosing widget to
		##   flash to the last size set by 'wm geometry.'
	    }
	}
	# Force display of watch cursor
	update idletasks
	set _watch_cursor_count 1
    } else {
	incr _watch_cursor_count
    }
    # Make sure watch cursor is removed no later than first idle
    # point after 30 seconds.
    after cancel $_watch_cursor_safetyid
    set _watch_cursor_safetyid [after 30000 Ow_PopWatchCursor 1]
}
proc Ow_PopWatchCursor { {force 0} } {
    global _watch_cursor_count _watch_cursor_safetyid
    if {$force || $_watch_cursor_count<=1} {
	# Last watch cursor set event.
	. config -cursor {}
	foreach win [winfo children .] {
	    if {[string compare $win [winfo toplevel $win]]==0 && \
		    [string compare Menu [winfo class $win]]!=0} {
		$win config -cursor {}
		## See "Menu" note in Ow_PushWatchCursor.
	    }
	}
	after cancel $_watch_cursor_safetyid
	set _watch_cursor_count 0
    } else {
	incr _watch_cursor_count -1
    }
}



################################
# Modal message dialog box
#
# Note: Tk built-in bitmaps include error, info, question,
#  questhead and warning.
set ow(dialog,width) 12c
set ow(modal_dialog,instance) 0
proc Ow_ModalDialog { title bitmap message {width ""} {scrollbars 0} \
        {defaultbtn ""} args } {
    global ow
    set instance $ow(modal_dialog,instance)
    incr ow(modal_dialog,instance)
    if {"$width"==""} {set width $ow(dialog,width)}
    if {"$scrollbars"==""} {set scrollbars 0}
    set window [toplevel .owModalDialog$instance]
    wm group $window .
    wm title $window "$title"
    Ow_PositionChild $window
 
    frame $window.top    -bd 10
    frame $window.bottom -bd 10
    if { "$bitmap" != "" } then {
        label $window.top.bitmap -bitmap $bitmap
        pack $window.top.bitmap -side left -anchor w -padx 2m
    }
    message $window.top.msg -text "$message"
    $window.top.msg configure -width $width
    pack $window.top.msg -side left -fill both -expand 1
    set btncount 0
    if {[info exists ow(modal_dialog,btncode$instance)]} {
        unset ow(modal_dialog,btncode$instance)  ;# Safety
    }
    foreach btnlabel $args {
        set btn($btncount) [button $window.bottom.btn$btncount \
                -text $btnlabel \
                -command "set ow(modal_dialog,btncode$instance) $btncount"]
        bind $btn($btncount) <Key-Return> "$btn($btncount) invoke"
        pack $btn($btncount) -side left -expand 1 -padx 5
        incr btncount 1
    }
    if {$btncount==0} {
        set btn($btncount) [button $window.bottom.btn$btncount -text "OK" \
                -command "set ow(modal_dialog,btncode$instance) $btncount"]
        pack $btn($btncount) -side left -expand 1 -padx 5
        incr btncount 1
    }
    wm protocol $window WM_DELETE_WINDOW \
            "set ow(modal_dialog,btncode$instance) 0"
    pack $window.top $window.bottom -side top -fill both -expand 1
 
    if {"$defaultbtn"!="" && $defaultbtn>=0 && $defaultbtn<$btncount} {
        # Code to make a "default" frame around default button.
        set fout [frame $window.bottom.defaultouter -bd 2 -relief sunken]
        lower $fout
        pack $fout -before $btn($defaultbtn) \
                -side left -expand 1 -padx 5
        pack forget $btn($defaultbtn)
        pack $btn($defaultbtn) -in $fout -padx 4 -pady 4
        # Give this button the keyboard focus
        focus $btn($defaultbtn)
    }
 
    update idletasks
    focus $window
    catch {tkwait visibility $window; grab $window }
    # Note: It appears dangerous to put a "grab" on a window
    #  before it is visible.  (Errant mouse clicks???)
    # Also, the sequence (including order) of
    #  'update idletasks ; focus $window ; catch ...'
    # seems to protect against binding reentrancy from the
    # user clicking too many times, or hitting keys too fast.
 
    Ow_SetIcon $window
    catch {tkwait variable ow(modal_dialog,btncode$instance) }
    set returncode $ow(modal_dialog,btncode$instance)
    unset ow(modal_dialog,btncode$instance)
    grab release $window
    destroy $window
    return $returncode
}

###################################################
### Nonfatal warning message dialog box (modal) ###
###################################################
proc Ow_NonFatalError { msg {title {}} {exitcode 1} } {
    if {"$title"==""} {
        global ProgName
        if [info exists ProgName] { set title "$ProgName: " }
        append title "Nonfatal Error"
    }
    set usercode [Ow_ModalDialog $title warning $msg {} {} 0 \
            "Continue" "Stack" "Die"]
    if {$usercode==2} { exit $exitcode }
    if {$usercode==1} {
        global errorCode errorInfo
        set msg "${errorInfo}\n-----------\nAddition info: $errorCode"
        set usercode [Ow_ModalDialog "${title}-Stack" info $msg {} {} 0 \
                "Continue" "Die"]
        if {$usercode==1} { exit $exitcode }
    }
    return 0
}

################################
# Modal propmt dialog box
#   Returns a 2 element list.  The first element is the return
# string, the second element is either 0 or 1; 0 indicates
# an "OK" return, 1 implies a "Cancel" return.  In the latter
# case the "defaultvalue" will always be returned as the first
# element
#
proc Ow_PromptDialog { title message {defaultvalue {}} \
        {width {}} {checkscript {}} } {
    # "checkscript" not yet implemented
    global ow
    if {[string match {} $width]} {set width $ow(dialog,width)}
    set window [toplevel .owPromptDialog]
    wm group $window .
    wm title $window "$title"
    Ow_PositionChild $window

    frame $window.top    -bd 10
    frame $window.bottom -bd 10
    set msg [message $window.top.msg -text $message -width $width]
    set value [entry $window.top.entry]
    $value insert 0 $defaultvalue

    pack $msg -side top -fill both -expand 1
    pack $value -side top -fill x -expand 1

    set btnOK [button $window.bottom.btnOK -text "OK" \
            -command "set ow(prompt_dialog,btn) 0"]
    set btnCancel [button $window.bottom.btnCancel -text "Cancel" \
            -command "set ow(prompt_dialog,btn) 1"]
    pack $btnOK $btnCancel -side left -expand 1 -padx 5

    wm protocol $window WM_DELETE_WINDOW \
            "set ow(prompt_dialog,btn) 0"

    pack $window.top $window.bottom -side top -fill both -expand 1

    # Extra bindings
    bind $value <Key-Return> "$btnOK invoke"
    bind $value <Key-Escape> \
            "$value delete 0 end ; $value insert 0 \{$defaultvalue\}"
    bind $btnOK <Key-Return> "$btnOK invoke"
    bind $btnCancel <Key-Return> "$btnCancel invoke"
 
    $value selection range 0 end
    focus $value
 
    update idletasks
    focus $window
    catch {tkwait visibility $window; grab $window }
    # Note: It appears dangerous to put a "grab" on a window
    #  before it is visible.  (Errant mouse clicks???)
    # Also, the sequence (including order) of
    #  'update idletasks ; focus $window ; catch ...'
    # seems to protect against binding reentrancy from the
    # user clicking too many times, or hitting keys too fast.
 
    Ow_SetIcon $window
    catch {tkwait variable ow(prompt_dialog,btn) }
    if {$ow(prompt_dialog,btn)==0} {
        # OK invoked
        set result [list [$value get] 0]
    } else {
        # Cancel invoked
        set result [list $defaultvalue 1]
    }    
    unset ow(prompt_dialog,btn)
 
    grab release $window
    destroy $window
    return $result
}

proc Ow_ModalMessage {msg type src} {

    if {[catch {winfo exists .} result] || !$result} {
        Oc_DefaultMessage $msg $type $src
        return
    }

    switch $type {
        panic {
            set bitmap error
        }
        error {
            set bitmap error
        }
        warning {
            set bitmap warning
        }
        default {
            set bitmap info
        }
    }

#    Ow_ModalDialog "[pid] $src $type" $bitmap $msg
    Ow_NonFatalError $msg "[pid] $src $type" 1

}

Oc_Log SetLogHandler Ow_ModalMessage panic Oc_Log
Oc_Log SetLogHandler Ow_ModalMessage error Oc_Log
Oc_Log SetLogHandler Ow_ModalMessage warning Oc_Log
Oc_Log SetLogHandler Ow_ModalMessage panic
Oc_Log SetLogHandler Ow_ModalMessage error
Oc_Log SetLogHandler Ow_ModalMessage warning

